import cgitb as _cgitb
import json as _json
import logging as _log
import unittest.mock as _mock
import pathlib as _pl

from PyQt5 import QtWidgets as _widgets

import trnsysGUI.StorageTank as _st

# Sometimes PyQT crashes only returning with quite a cryptic error code. Sometimes, again, we can get
# a more helpful stack trace using the cgitb module.
_cgitb.enable(format="text")


class _StrictMock:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TestStorageTank:
    LEGACY_JSON_PATH = _pl.Path(__file__).parent / "data" / "storageTankOldestFormat.json"

    def testDeserializeJsonFromLegacyFormatAndSerialize(self, tmp_path):  # pylint: disable=invalid-name
        expectedPath = _pl.Path(__file__).parent / "data" / "storageTankNewestFormat.json"
        expectedStorageTankJson = expectedPath.read_text()

        logger = _log.getLogger("root")
        (
            diagramViewMock,
            objectsNeededToBeKeptAliveWhileTanksAlive,  # pylint: disable=unused-variable
        ) = self._createDiagramViewMocksAndOtherObjectsToKeepAlive(logger, tmp_path)

        legacyJson = self.LEGACY_JSON_PATH.read_text()
        storageTank = self._deserializeStorageTank(legacyJson, diagramViewMock)

        serializedStorageTank = storageTank.encode()[1]
        actualStorageTankJson = _json.dumps(
            serializedStorageTank, indent=4, sort_keys=True
        )

        assert actualStorageTankJson == expectedStorageTankJson

        self._deserializeStorageTank(actualStorageTankJson, diagramViewMock)

    @staticmethod
    def _deserializeStorageTank(storageTankLegacyJson, diagramViewMock):
        legacySerializedStorageTank = _json.loads(storageTankLegacyJson)

        storageTank = _st.StorageTank(trnsysType="StorageTank", parent=diagramViewMock)  # pylint: disable=no-member
        diagramViewMock.scene().addItem(storageTank)

        blocks = []
        storageTank.decode(legacySerializedStorageTank, blocks)

        return storageTank

    def testExportDdck(self, tmp_path):  # pylint: disable=invalid-name
        logger = _log.getLogger("root")
        (
            diagramViewMock,
            objectsNeededToBeKeptAliveWhileTanksAlive,  # pylint: disable=unused-variable
        ) = self._createDiagramViewMocksAndOtherObjectsToKeepAlive(logger, tmp_path)

        legacyJson = self.LEGACY_JSON_PATH.read_text()
        storageTank = self._deserializeStorageTank(legacyJson, diagramViewMock)

        self._setupExternalConnectionMocks(storageTank)

        storageTank.exportDck()

        actualDdckPath = tmp_path / "ddck" / "StorageTank_7701" / "Dhw.ddck"
        actualDdckContent = actualDdckPath.read_text()
        print(actualDdckContent)

        expectedDdckContent = r"""*************************************
**BEGIN Dhw.ddck
*************************************

*************************************
** Plug-Flow Model exported from TRNSYS GUI
*************************************

*************************************
** To be checked: 
** check cp and rho values for the cirquits 
** default is cpwat and rhowat, for solarcirc usually cpbri and rhobri have to be used 
*************************************

*************************************
** outputs to energy balance in kWh
** Following this naming standard : qSysIn_name, qSysOut_name, elSysIn_name, elSysOut_name
*************************************
EQUATIONS 3
qSysOut_TesDhwLoss = sumQLoss_Tes7703
qSysOut_TesDhwAcum = sumQAcum_Tes7703
elSysIn_Q_TesDhwAux = qHeatSource_Tes7703

*************************************
*** Inputs from hydraulic solver ****
*************************************
EQUATIONS 3
Tdp1In_Tes7703 = Tdpp0ExtFromPortConn ! @connectDdck
Mfrdp1_Tes7703 = Mfrdpp0ExtFromPortConn ! @connectDdck
Tdp1InRev_Tes7703 = Tdpp0ExtToPortConn ! @connectDdck
EQUATIONS 3
Tdp2In_Tes7703 = Tdpp1ExtFromPortConn ! @connectDdck
Mfrdp2_Tes7703 = Mfrdpp1ExtFromPortConn ! @connectDdck
Tdp2InRev_Tes7703 = Tdpp1ExtToPortConn ! @connectDdck
EQUATIONS 3
Tdp3In_Tes7703 = Tdpp2ExtFromPortConn ! @connectDdck
Mfrdp3_Tes7703 = Mfrdpp2ExtFromPortConn ! @connectDdck
Tdp3InRev_Tes7703 = Tdpp2ExtToPortConn ! @connectDdck
EQUATIONS 3
Thx1In_Tes7703 = Thx0ExtFromPortConn ! @connectDdck
Mfrhx1_Tes7703 = Mfrhx0ExtFromPortConn ! @connectDdck
Thx1InRev_Tes7703 = Thx0ExtToPortConn ! @connectDdck
CONSTANTS 2
zAux1_Tes7703=0.00 !connectDDck
qAux1_Tes7703=0.00 !connectDDck

*************************************
**** Outputs to hydraulic solver ****
*************************************
EQUATIONS 3
*** direct port outputs
Tdp1Out_Tes7703=[50,1] ! 
Tdp2Out_Tes7703=[50,3] ! 
Tdp3Out_Tes7703=[50,5] ! 
EQUATIONS 1
*** heat exchanger outputs
Thx1Out_Tes7703=[50,102] ! 

*************************************
****** Parameters of Type1924 *******
*************************************
CONSTANTS 1
ratioTes7703 = Vol_Tes7703 / VStoreRef
*********Connecting values of DIRECT PORT=1***********
CONSTANTS 2
zInDp1_Tes7703=0.95 ! @connectDdck 
zOutDp1_Tes7703=0.35 ! @connectDdck 
*********Connecting values of DIRECT PORT=2***********
CONSTANTS 2
zInDp2_Tes7703=0.70 ! @connectDdck 
zOutDp2_Tes7703=0.90 ! @connectDdck 
*********Connecting values of DIRECT PORT=3***********
CONSTANTS 2
zInDp3_Tes7703=0.05 ! @connectDdck 
zOutDp3_Tes7703=0.95 ! @connectDdck 
*********Connecting values of HX=1***********
CONSTANTS 4
zInhx1_Tes7703=0.30  ! @connectDdck
zOuthx1_Tes7703=0.12 ! @connectDdck
Cphx1_Tes7703=cpwat     ! @connectDdck
Rhohx1_Tes7703=rhowat    ! @connectDdck
*********Constant of DIRECT PORT=1***********
CONSTANTS 1
Dp1Strat_Tes7703=0 ! 0: no forced stratification ; 1: force to stratify
*********Constant of DIRECT PORT=2***********
CONSTANTS 1
Dp2Strat_Tes7703=0 ! 0: no forced stratification ; 1: force to stratify
*********Constant of DIRECT PORT=3***********
CONSTANTS 1
Dp3Strat_Tes7703=0 ! 0: no forced stratification ; 1: force to stratify
********** HEAT EXCHANGER CONSTANTS*******
CONSTANTS 1
nHxUsed_Tes7703=1 
*********Constant values of HX=1***********
CONSTANTS 15
dInHx1_Tes7703=0.05 ! m only if modHx1=1
dOutHx1_Tes7703=0.06 ! m only if modHx1=1
LHx1_Tes7703=20 ! m only if modHx1=1
LamHx1_Tes7703=50 ! W/mK only if modHx1=1
brineConcHx1_Tes7703=30 ! [0-100] only if modHx1=1
VHx1_Tes7703=0. 
nCvHx1_Tes7703=20 
modHx1_Tes7703=1 ! modHX1, 0 = physical model, 1 = Drueck-model (Multiport)
nNuHx1_Tes7703=0.5 !  only if modHx1=1
cNuHx1_Tes7703=0.25 !  only if modHx1=1
dUaMfrHx1_Tes7703=0.375 ! only if modHx1=0
dUadTHx1_Tes7703=0.0 ! only if modHx1=0
dUaTHx1_Tes7703=0.458 ! only if modHx1=0
UaHx1_Tes7703=1105*ratioTes7703 ! kJ/hK only if modHx1=0
startUpHx1_Tes7703=0. ! only if modHx1=0
*************************************
** USER DEFINED TEMPERATURE SENSOR HEIGHTS. To be changed by user 
*************************************
CONSTANTS 10
zSen1_Tes7703=0.05
zSen2_Tes7703=0.15
zSen3_Tes7703=0.25
zSen4_Tes7703=0.35
zSen5_Tes7703=0.45
zSen6_Tes7703=0.55
zSen7_Tes7703=0.65
zSen8_Tes7703=0.75
zSen9_Tes7703=0.85
zSen10_Tes7703=0.95
*************************************
** USER DEFINED AVERAGED TEMPERATURE SENSOR HEIGHTS. To be changed by user 
*************************************
CONSTANTS 10
zSenAvgBot1_Tes7703 = 0.05
zSenAvgTop1_Tes7703 = 0.15
zSenAvgBot2_Tes7703 = 0.25
zSenAvgTop2_Tes7703 = 0.35
zSenAvgBot3_Tes7703 = 0.45
zSenAvgTop3_Tes7703 = 0.55
zSenAvgBot4_Tes7703 = 0.65
zSenAvgTop4_Tes7703 = 0.75
zSenAvgBot5_Tes7703 = 0.85
zSenAvgTop5_Tes7703 = 0.95
CONSTANTS 12
Vol_Tes7703=1          ! 1: m3, volume of store
RhoWat_Tes7703=RhoWat  ! 2: kg/m3, density of storage media
CpWat_Tes7703=CpWat    ! 3: kJ/kgK, specific heat of storage media
lamZ_Tes7703=0.6       ! 4: W/mK, effective vertical thermal conductivity of TES
Heigh_Tes7703=1.       ! 5: m, storage height
Tini_Tes7703=60.       ! 6: oC, initial temperature
nCvMax_Tes7703=400     ! 7: -, minimum relative plug height
nCvMin_Tes7703=20      ! 8: -, maximum relative plug height
maxTDiff_Tes7703=0.015 ! 9: K, maximum temperature difference between plugs
readMode_Tes7703=0     ! 10: 1: from table, 0: Tini and CapTot
Tref_Tes7703=273.15    ! 11: oC, reference temperature
Tmax_Tes7703=100.       ! 6: oC, initial temperature
CONSTANTS 10
Ufoam_Tes7703= 0.67 ! W/(m2K) 6 cm of foam of 0.04 W/(mK) 
Ubot_Tes7703 = 1.5 ! W/(m2K) 2 cm of foam of 0.04 W/(mK)
Atop_Tes7703 = Vol_Tes7703/Heigh_Tes7703 ! m2
Diameter_Tes7703 = (4*ATop_Tes7703/PI)^0.5 ! m 
ALat_Tes7703 = Heigh_Tes7703*PI*Diameter_Tes7703 ! m2
UaBot_Tes7703= Ubot_Tes7703*ATop_Tes7703 ! @userDefined W/k 
Uaz1_Tes7703 = Ufoam_Tes7703*ALat_Tes7703/3 ! @userDefined W/k
Uaz2_Tes7703 = Ufoam_Tes7703*ALat_Tes7703/3 ! @userDefined W/k
Uaz3_Tes7703 = Ufoam_Tes7703*ALat_Tes7703/3 ! @userDefined W/k
UaTop_Tes7703 = Ufoam_Tes7703*ATop_Tes7703 ! @userDefined W/k
************* MOVING PLATE *******************
CONSTANTS 4
MoInsPlate_Tes7703=0 ! 0-2, Insulation Plate Mode: 0 = no insulation plate inside TES, 1 = insulation plate at fixed relative height, 2 = insulation plate at fixed temperature / density controlled 
zInsPlate_Tes7703=0  ! 0-1, relative position of fixed height insulation plate inside TES (only for Insulation Plate Mode = 1
TinsPlate_Tes7703=0  ! oC, temperature at which moveable insulation plate floats in TES (only for Insulation Plate Mode = 2)
UAinsPlate_Tes7703=0 ! W/K, overall heat transfer coefficient across moveable insulation plate (including heat transfer in gap between plate and wall and in wall at the respective height)
*************************************
********** TYPE DEFINITION **********
*************************************
UNIT 50 TYPE 7704     ! plug flow tank
PARAMETERS 219 
Vol_Tes7703     ! 1: m3, volume of store
RhoWat_Tes7703  ! 2: kg/m3, density of storage media
CpWat_Tes7703   ! 3: kJ/kgK, specific heat of storage media
lamZ_Tes7703    ! 4: W/mK, effective vertical thermal conductivity of TES
Heigh_Tes7703   ! 5: m, storage height
TIni_Tes7703   ! 6: oC, initial temperature
nCvMax_Tes7703  ! 7: -, minimum relative plug height
nCvMin_Tes7703  ! 8: -, maximum relative plug height
maxTDiff_Tes7703  ! 9: K, maximum temperature difference between plugs
readMode_Tes7703  ! 10: 1: from table, 0: Tini and CapTot
Tref_Tes7703     ! 11: oC, reference temperature
UaBot_Tes7703 ! W/k 
Uaz1_Tes7703  ! W/k
Uaz2_Tes7703  ! W/k
Uaz3_Tes7703  ! W/k
UaTop_Tes7703 ! W/k
tMax_Tes7703
0	0	0 ! 17-20 unsused parameters
zInDp1_Tes7703 zOutDp1_Tes7703 zero Dp1Strat_Tes7703 ! 21 - 25: zIn, zOut, cp, strat
zInDp2_Tes7703 zOutDp2_Tes7703 zero Dp2Strat_Tes7703 ! 26 - 30: zIn, zOut, cp, strat
zInDp3_Tes7703 zOutDp3_Tes7703 zero Dp3Strat_Tes7703 ! 31 - 35: zIn, zOut, cp, strat
-1 -1 zero zero ! 36 - 40: zIn, zOut, cp, strat
-1 -1 zero zero ! 41 - 45: zIn, zOut, cp, strat
-1 -1 zero zero ! 46 - 50: zIn, zOut, cp, strat
-1 -1 zero zero ! 51 - 55: zIn, zOut, cp, strat
-1 -1 zero zero ! 56 - 60: zIn, zOut, cp, strat
-1 -1 zero zero ! 61 - 65: zIn, zOut, cp, strat
-1 -1 zero zero ! 66 - 70: zIn, zOut, cp, strat
zSen1_Tes7703 zSen2_Tes7703 zSen3_Tes7703 zSen4_Tes7703 zSen5_Tes7703 zSen6_Tes7703 zSen7_Tes7703 zSen8_Tes7703 zSen9_Tes7703 zSen10_Tes7703 ! 61-71 : relative storage temperature sensor heights 
zSenAvgBot1_Tes7703 zSenAvgTop1_Tes7703 ! 71-72 : relative position of lower and upper edge temeprature sensors
zSenAvgBot2_Tes7703 zSenAvgTop2_Tes7703 ! 73-74 : relative position of lower and upper edge temeprature sensors
zSenAvgBot3_Tes7703 zSenAvgTop3_Tes7703 ! 75-76 : relative position of lower and upper edge temeprature sensors
zSenAvgBot4_Tes7703 zSenAvgTop4_Tes7703 ! 77-78 : relative position of lower and upper edge temeprature sensors
zSenAvgBot5_Tes7703 zSenAvgTop5_Tes7703 ! 79-80 : relative position of lower and upper edge temeprature sensors
MoInsPlate_Tes7703 ! 81: 0-2, Insulation Plate Mode: 0 = no insulation plate inside TES, 1 = insulation plate at fixed relative height, 2 = insulation plate at fixed temperature / density controlled 
zInsPlate_Tes7703  ! 82: 0-1, relative position of fixed height insulation plate inside TES (only for Insulation Plate Mode = 1
TinsPlate_Tes7703  ! 83: oC, temperature at which moveable insulation plate floats in TES (only for Insulation Plate Mode = 2)
UAinsPlate_Tes7703 ! 84: W/K, overall heat transfer coefficient across moveable insulation plate (including heat transfer in gap between plate and wall and in wall at the respective height)
nHxUsed_Tes7703     ! 85: number Of used Hx
** Parameters for heat Exchanger number 1
zInHx1_Tes7703 zOutHx1_Tes7703 dInHx1_Tes7703 dOutHx1_Tes7703 LHx1_Tes7703 LamHx1_Tes7703 brineConcHx1_Tes7703 VHx1_Tes7703 CpHx1_Tes7703 RhoHx1_Tes7703 nCvHx1_Tes7703 modHx1_Tes7703 nNuHx1_Tes7703 cNuHx1_Tes7703 dUaMfrHx1_Tes7703 dUadTHx1_Tes7703 dUaTHx1_Tes7703 UaHx1_Tes7703 startUpHx1_Tes7703 ! Heax exchanger 1
** Parameters for heat Exchanger number 2
-1 -1 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero ! Heax exchanger 2
** Parameters for heat Exchanger number 3
-1 -1 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero ! Heax exchanger 3
** Parameters for heat Exchanger number 4
-1 -1 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero ! Heax exchanger 4
** Parameters for heat Exchanger number 5
-1 -1 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero ! Heax exchanger 5
** Parameters for heat Exchanger number 6
-1 -1 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero ! Heax exchanger 6
** 20 height position for any heat source, e.g. electrical backup or heat pump condenser. Any position can be due to a different heat source
zAux1_Tes7703 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero 
INPUTS 69
************10 DIRECT PORTS INPUTS***************
Tdp1In_Tes7703 Mfrdp1_Tes7703 Tdp1InRev_Tes7703
Tdp2In_Tes7703 Mfrdp2_Tes7703 Tdp2InRev_Tes7703
Tdp3In_Tes7703 Mfrdp3_Tes7703 Tdp3InRev_Tes7703
zero zero zero
zero zero zero
zero zero zero
zero zero zero
zero zero zero
zero zero zero
zero zero zero
****************
TroomStore
***************** 6 HX INPUTS ******************
Thx1In_Tes7703 Mfrhx1_Tes7703 Thx1InRev_Tes7703
zero zero zero
zero zero zero
zero zero zero
zero zero zero
zero zero zero
***************** 20 HEAT SOURCE INPUTS ******************
qAux1_Tes7703 zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero zero 
****************** INTIAL INPUTS***********************
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero zero 
zero zero zero zero zero zero zero zero zero 
*****************OUTPUTS****************
EQUATIONS 3
Qdp1_Tes7703=[50,31] ! 
Qdp2_Tes7703=[50,32] ! 
Qdp3_Tes7703=[50,33] ! 
EQUATIONS 21
TAvg_Tes7703 = [50,180] ! Average storage temperature 
***Temperatures at 10 equallay distributed height 
T1_Tes7703 =[50,21] !temperature at 0.05 
T2_Tes7703 =[50,22] !temperature at 0.15 
T3_Tes7703 =[50,23] !temperature at 0.25 
T4_Tes7703 =[50,24] !temperature at 0.35 
T5_Tes7703 =[50,25] !temperature at 0.45 
T6_Tes7703 =[50,26] !temperature at 0.55 
T7_Tes7703 =[50,27] !temperature at 0.65 
T8_Tes7703 =[50,28] !temperature at 0.75 
T9_Tes7703 =[50,29] !temperature at 0.85 
T10_Tes7703 =[50,30] !temperature at 0.95 
***Temperatures at 10 sensors user defined height
Tsen1_Tes7703 =[50,71] ! temperature at user defined sensor height Tsen1_Tes7703 
Tsen2_Tes7703 =[50,72] ! temperature at user defined sensor height Tsen2_Tes7703 
Tsen3_Tes7703 =[50,73] ! temperature at user defined sensor height Tsen3_Tes7703 
Tsen4_Tes7703 =[50,74] ! temperature at user defined sensor height Tsen4_Tes7703 
Tsen5_Tes7703 =[50,75] ! temperature at user defined sensor height Tsen5_Tes7703 
Tsen6_Tes7703 =[50,76] ! temperature at user defined sensor height Tsen6_Tes7703 
Tsen7_Tes7703 =[50,77] ! temperature at user defined sensor height Tsen7_Tes7703 
Tsen8_Tes7703 =[50,78] ! temperature at user defined sensor height Tsen8_Tes7703 
Tsen9_Tes7703 =[50,79] ! temperature at user defined sensor height Tsen9_Tes7703 
Tsen10_Tes7703 =[50,80] ! temperature at user defined sensor height Tsen10_Tes7703 
EQUATIONS 1
Qhx1Out_Tes7703=[50,104] ! 
EQUATIONS 1
qHeatSource_Tes7703 = [50,181] ! Heat input of all auxiliary heat sources [kW]
EQUATIONS 5
sumQv_Tes7703     = [50,176] ! Heat input of all heat exchangers [kW]
sumQLoss_Tes7703  = [50,177] ! Heat Losses of the Tes [kW]
sumQAcum_Tes7703  = [50,178] ! Sensible accumulated heat [kW]
sumQPorts_Tes7703 = [50,179] ! Heat Input by direct ports [kW]
Imb_Tes7703       = [50,64]  ! Heat Imbalance in Tes  IMB = sumQv - sumQLoss -sumQAcum + sumQPort
CONSTANTS 1 
unitPrinter_Tes7703 = 51 
ASSIGN temp\TES7703_MO.Prt unitPrinter_Tes7703
UNIT 51 TYPE 46
PARAMETERS 5
unitPrinter_Tes7703 ! 1: Logical unit number, -
-1  ! 2: Logical unit for monthly summaries
1 ! 3: Relative or absolute start time. 0: print at time intervals relative to the simulation start time. 1: print at absolute time intervals. No effect for monthly integrations
-1  ! 4: Printing & integrating interval, h. -1 for monthly integration
0  ! 5: Number of inputs to avoid integration
INPUTS 10
sumQv_Tes1 sumQLoss_Tes1 sumQAcum_Tes1 sumQPorts_Tes1 Imb_Tes1 Qdp1_Tes7703 Qdp2_Tes7703 Qdp3_Tes7703 Qhx1Out_Tes7703 qHeatSource_Tes7703 
zero zero zero zero zero zero zero zero zero zero 
*************************************
********** Online Plotter ***********
*************************************

UNIT 501 TYPE 65     ! Online Plotter HX 
PARAMETERS 12   
10     ! 1 Nb. of left-axis variables 
0     ! 2 Nb. of right-axis variables
0     ! 3 Left axis minimum 
100     ! 4 Left axis maximum -
0     ! 5 Right axis minimum 
100     ! 6 Right axis maximum 
nPlotsPerSim     ! 7 Number of plots per simulation 
12     ! 8 X-axis gridpoints
1     ! 9 Shut off Online w/o removing 
-1     ! 10 Logical unit for output file 
0     ! 11 Output file units
0     ! 12 Output file delimiter
INPUTS 10     
T1_Tes7703 T2_Tes7703 T3_Tes7703 T4_Tes7703 T5_Tes7703 T6_Tes7703 T7_Tes7703 T8_Tes7703 T9_Tes7703 T10_Tes7703 
T1_Tes7703 T2_Tes7703 T3_Tes7703 T4_Tes7703 T5_Tes7703 T6_Tes7703 T7_Tes7703 T8_Tes7703 T9_Tes7703 T10_Tes7703 
LABELS  3         
Temperatures  
MassFlows  
Tes%d 


"""

        assert actualDdckContent == expectedDdckContent

    @staticmethod
    def _setupExternalConnectionMocks(storageTank):
        for i, heatExchanger in enumerate(storageTank.heatExchangers):
            externalFromPortConnection = _StrictMock(displayName=f"hx{i}ExtFromPortConn")
            heatExchanger.port1.connectionList.append(externalFromPortConnection)

            externalToPortConnection = _StrictMock(displayName=f"hx{i}ExtToPortConn")
            heatExchanger.port2.connectionList.append(externalToPortConnection)

        for i, directPortPair in enumerate(storageTank.directPortPairs):
            externalFromPortConnection = _StrictMock(
                displayName=f"dpp{i}ExtFromPortConn",
                fromPort=_StrictMock(),
                toPort=directPortPair.fromPort,
            )
            directPortPair.fromPort.connectionList.append(
                externalFromPortConnection
            )

            externalToPortConnection = _StrictMock(
                displayName=f"dpp{i}ExtToPortConn",
                fromPort=directPortPair.toPort,
                toPort=_StrictMock(),
            )
            directPortPair.toPort.connectionList.append(
                externalToPortConnection
            )

    @staticmethod
    def _createDiagramViewMocksAndOtherObjectsToKeepAlive(logger, projectPath):
        application = _widgets.QApplication([])

        mainWindow = _widgets.QMainWindow()

        editorMock = _widgets.QWidget(parent=mainWindow)
        editorMock.connectionList = []
        editorMock.logger = logger
        editorMock.trnsysObj = []
        editorMock.groupList = []
        editorMock.projectPath = str(projectPath)
        editorMock.splitter = _mock.Mock(name="splitter")
        editorMock.idGen = _mock.Mock(
            name="idGen",
            spec_set=[
                "getID",
                "getTrnsysID",
                "getStoragenTes",
                "getStorageType",
                "getConnID",
            ],
        )
        editorMock.moveDirectPorts = True
        editorMock.editorMode = 1
        editorMock.snapGrid = False
        editorMock.alignMode = False

        editorMock.idGen.getID = lambda: 7701
        editorMock.idGen.getTrnsysID = lambda: 7702
        editorMock.idGen.getStoragenTes = lambda: 7703
        editorMock.idGen.getStorageType = lambda: 7704
        editorMock.idGen.getConnID = lambda: 7705

        graphicsScene = _widgets.QGraphicsScene(parent=editorMock)
        editorMock.diagramScene = graphicsScene

        diagramViewMock = _widgets.QGraphicsView(graphicsScene, parent=editorMock)
        diagramViewMock.logger = logger

        mainWindow.setCentralWidget(editorMock)
        mainWindow.showMinimized()

        return diagramViewMock, [application, mainWindow, editorMock, graphicsScene]



# pylint: skip-file
# type: ignore

import os
import pathlib as _pl
import shutil
import subprocess
import sys

from PyQt5.QtWidgets import *
from pytrnsys.utils import log

import trnsysGUI.arguments as args
import trnsysGUI.buildDck as dckBuilder
import trnsysGUI.common.cancelled as _ccl
import trnsysGUI.common.error as _err
import trnsysGUI.diagram.Editor as _de
import trnsysGUI.images as _img
import trnsysGUI.project as _prj
import trnsysGUI.settings as _settings
import trnsysGUI.settingsDlg as _sdlg
import trnsysGUI.tracing as trc
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.Connection import Connection
from trnsysGUI.Graphicaltem import GraphicalItem
from trnsysGUI.MassFlowVisualizer import MassFlowVisualizer
from trnsysGUI.ProcessMain import ProcessMain
from trnsysGUI.RunMain import RunMain
from trnsysGUI.StorageTank import StorageTank
from trnsysGUI.configFile import configFile

__version__ = "1.0.0"
__author__ = "Stefano Marti"
__email__ = "stefano.marti@spf.ch"
__status__ = "Prototype"


class _MainWindow(QMainWindow):
    def __init__(self, logger, project: _prj.Project, parent=None):
        super().__init__(parent)

        self.jsonPath = None
        self.logger = logger

        self.centralWidget = self._createDiagramEditor(project)
        self.setCentralWidget(self.centralWidget)

        self.labelVisState = False
        self.massFlowEnabled = False
        self.calledByVisualizeMf = False
        self.currentFile = "Untitled"

        # Toolbar actions
        saveDiaAction = QAction(_img.INBOX_PNG.icon(), "Save", self)
        saveDiaAction.triggered.connect(self.saveDia)

        loadDiaAction = QAction(_img.OUTBOX_PNG.icon(), "Open", self)
        loadDiaAction.triggered.connect(self.loadDia)

        updateConfigAction = QAction(
            _img.UPDATE_CONFIG_PNG.icon(), "Update run.config", self
        )
        updateConfigAction.triggered.connect(self.updateRun)

        runSimulationAction = QAction(
            _img.RUN_SIMULATION_PNG.icon(), "Run simulation", self
        )
        runSimulationAction.triggered.connect(self.runSimulation)

        processSimulationAction = QAction(
            _img.PROCESS_SIMULATION_PNG.icon(), "Process data", self
        )
        processSimulationAction.triggered.connect(self.processSimulation)

        deleteDiaAction = QAction(_img.TRASH_PNG.icon(), "Delete diagram", self)
        deleteDiaAction.triggered.connect(self.deleteDia)

        zoomInAction = QAction(_img.ZOOM_IN_PNG.icon(), "Zoom in", self)
        zoomInAction.triggered.connect(self.setZoomIn)

        zoomOutAction = QAction(_img.ZOOM_OUT_PNG.icon(), "Zoom out", self)
        zoomOutAction.triggered.connect(self.setZoomOut)

        toggleConnLabels = QAction(_img.LABEL_TOGGLE_PNG.icon(), "Toggle labels", self)
        toggleConnLabels.triggered.connect(self.toggleConnLabels)

        exportHydraulicsAction = QAction(
            _img.EXPORT_HYDRAULICS_PNG.icon(), "Export hydraulic.ddck", self
        )
        exportHydraulicsAction.triggered.connect(self.exportHydraulicsDdck)

        exportHydCtrlAction = QAction(
            _img.EXPORT_HYDRAULIC_CONTROL_PNG.icon(),
            "Export hydraulic_control.ddck",
            self,
        )
        exportHydCtrlAction.triggered.connect(self.exportHydraulicControl)

        exportDckAction = QAction(_img.EXPORT_DCK_PNG.icon(), "Export dck", self)
        exportDckAction.triggered.connect(self.exportDck)

        editGroupsAction = QAction("Edit groups/loops", self)
        editGroupsAction.triggered.connect(self.editGroups)

        selectMultipleAction = QAction("Select multiple items", self)
        selectMultipleAction.triggered.connect(self.createSelection)
        selectMultipleAction.setShortcut("s")

        toggleSnapAction = QAction("Toggle snap grid", self)
        toggleSnapAction.triggered.connect(self.toggleSnap)
        toggleSnapAction.setShortcut("a")

        toggleAlignModeAction = QAction("Toggle align mode", self)
        toggleAlignModeAction.triggered.connect(self.toggleAlignMode)
        toggleAlignModeAction.setShortcut("q")

        runMassflowSolverAction = QAction(
            _img.RUN_MFS_PNG.icon(), "Run the massflow solver", self
        )
        runMassflowSolverAction.triggered.connect(self.runAndVisMf)

        openVisualizerAction = QAction(
            _img.VIS_MFS_PNG.icon(), "Start visualization of mass flows", self
        )
        openVisualizerAction.triggered.connect(self.visualizeMf)

        trnsysList = QAction("Print trnsysObj", self)
        trnsysList.triggered.connect(self.mb_debug)

        loadVisual = QAction("Load MRF", self)
        loadVisual.triggered.connect(self.loadVisualization)

        runAction = QAction(_img.ROTATE_TO_RIGHT_PNG.icon(), "Run", self)
        runAction.triggered.connect(self.runApp)

        # Tool bar
        tb = self.addToolBar("Main Toolbar...")
        tb.setObjectName("Toolbar")
        tb.addAction(saveDiaAction)
        tb.addAction(loadDiaAction)
        tb.addAction(zoomInAction)
        tb.addAction(zoomOutAction)
        tb.addAction(toggleConnLabels)
        tb.addAction(runMassflowSolverAction)
        tb.addAction(openVisualizerAction)
        tb.addAction(exportHydraulicsAction)
        tb.addAction(exportHydCtrlAction)
        tb.addAction(updateConfigAction)
        tb.addAction(exportDckAction)
        tb.addAction(runSimulationAction)
        tb.addAction(processSimulationAction)
        tb.addAction(deleteDiaAction)

        # Menu bar actions
        self.fileMenu = QMenu("File")

        fileMenuNewAction = QAction("New", self)
        fileMenuNewAction.triggered.connect(self.newDia)
        fileMenuNewAction.setShortcut("Ctrl+n")
        self.fileMenu.addAction(fileMenuNewAction)

        fileMenuOpenAction = QAction("Open", self)
        fileMenuOpenAction.triggered.connect(self.openFile)
        fileMenuOpenAction.setShortcut("Ctrl+o")
        self.fileMenu.addAction(fileMenuOpenAction)

        fileMenuSaveAction = QAction("Save", self)
        fileMenuSaveAction.triggered.connect(self.saveDia)
        fileMenuSaveAction.setShortcut("Ctrl+s")
        self.fileMenu.addAction(fileMenuSaveAction)

        fileMenuCopyToNewAction = QAction("Copy to new folder", self)
        fileMenuCopyToNewAction.triggered.connect(self.copyToNew)
        self.fileMenu.addAction(fileMenuCopyToNewAction)

        exportAsPDF = QAction("Export as PDF", self)
        exportAsPDF.triggered.connect(self.exportPDF)
        exportAsPDF.setShortcut("Ctrl+e")
        self.fileMenu.addAction(exportAsPDF)

        debugConnections = QAction("Debug Conn", self)
        debugConnections.triggered.connect(self.debugConns)
        self.fileMenu.addAction(debugConnections)

        setTransysPath = QAction("Set TRNSYS path", self)
        setTransysPath.triggered.connect(self.askUserForSettingsValuesAndSave)
        self.fileMenu.addAction(setTransysPath)

        self.editMenu = QMenu("Edit")
        self.editMenu.addAction(toggleSnapAction)
        self.editMenu.addAction(toggleAlignModeAction)

        AboutAction = QAction("About", self)
        AboutAction.triggered.connect(self.showAbout)

        VersionAction = QAction("Version", self)
        VersionAction.triggered.connect(self.showVersion)

        CreditsAction = QAction("Credits", self)
        CreditsAction.triggered.connect(self.showCredits)

        self.helpMenu = QMenu("Help")
        self.helpMenu.addAction(AboutAction)
        self.helpMenu.addAction(VersionAction)
        self.helpMenu.addAction(CreditsAction)

        # Menu bar
        self.mb = self.menuBar()
        self.mb.addMenu(self.fileMenu)
        self.mb.addMenu(self.editMenu)
        self.mb.addMenu(self.helpMenu)
        self.mb.addSeparator()

        # Status bar
        self.sb = self.statusBar()
        self.sb.showMessage("Mode is " + str(self.centralWidget.editorMode))

        # QUndo framework
        self.undoStack = QUndoStack(self)
        undoAction = self.undoStack.createUndoAction(self, "Undo")
        undoAction.setShortcut("Ctrl+z")

        redoAction = self.undoStack.createRedoAction(self, "Redo")
        redoAction.setShortcut("Ctrl+y")

        self.editMenu.addAction(undoAction)
        self.editMenu.addAction(redoAction)

    def newDia(self):
        messageBox = QMessageBox()
        messageBox.setText(
            "Are you sure you want to start a new project? Unsaved progress on the current one will be lost."
        )
        messageBox.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        messageBox.setDefaultButton(QMessageBox.Cancel)

        result = messageBox.exec()
        if result == QMessageBox.Cancel:
            return

        startingDirectoryPath = _pl.Path(self.projectFolder).parent

        createProjectMaybeCancelled = _prj.getCreateProject(startingDirectoryPath)
        if _ccl.isCancelled(createProjectMaybeCancelled):
            return
        createProject = _ccl.value(createProjectMaybeCancelled)

        self.centralWidget = self._createDiagramEditor(createProject)
        self.setCentralWidget(self.centralWidget)

    def saveDia(self):
        self.logger.info("Saving diagram")
        self.centralWidget.save()

    def copyToNew(self):
        currentProjectFolderPath = _pl.Path(self.projectFolder)

        startingDirectoryPath = currentProjectFolderPath.parent

        maybeCancelled = _prj.getExistingEmptyDirectory(startingDirectoryPath)
        if _ccl.isCancelled(maybeCancelled):
            return
        newProjectFolderPath = _ccl.value(maybeCancelled)

        oldProjectFolderPath = _pl.Path(self.projectFolder)

        self._copyContents(oldProjectFolderPath, newProjectFolderPath)
        newJsonFilePath = self._changeAndGetNewJsonFileName(newProjectFolderPath, oldProjectFolderPath)

        loadProject = _prj.LoadProject(newJsonFilePath)

        self.centralWidget = self._createDiagramEditor(loadProject)
        self.setCentralWidget(self.centralWidget)
        self.centralWidget.save(showWarning=False)

    @staticmethod
    def _copyContents(oldProjectFolderPath, newProjectFolderPath):
        for child in oldProjectFolderPath.iterdir():
            destinationPath = newProjectFolderPath / child.name
            if child.is_dir():
                shutil.copytree(child, destinationPath)
            else:
                shutil.copy(child, destinationPath)

    @staticmethod
    def _changeAndGetNewJsonFileName(newProjectFolderPath, oldProjectFolderPath):
        oldJsonFileName = f"{oldProjectFolderPath.name}.json"
        newJsonFileName = f"{newProjectFolderPath.name}.json"
        shutil.move(
            newProjectFolderPath / oldJsonFileName,
            newProjectFolderPath / newJsonFileName,
        )
        newJsonFilePath = newProjectFolderPath / newJsonFileName
        return newJsonFilePath

    def loadDia(self):
        self.logger.info("Loading diagram")
        self.openFile()

    def updateRun(self):
        self.logger.info("Updating run.config")
        configPath = os.path.join(self.projectFolder, "run.config")
        runConfig = configFile(configPath, self.centralWidget)
        runConfig.updateConfig()

    def runSimulation(self):
        ddckPath = os.path.join(self.projectFolder, "ddck")

        #   Check hydraulic.ddck
        hydraulicPath = os.path.join(ddckPath, "hydraulic\\hydraulic.ddck")
        if not os.path.isfile(hydraulicPath):
            self.exportHydraulicsDdck()
        infile = open(hydraulicPath, "r")
        hydraulicLines = infile.readlines()
        blackBoxLines = []
        for i in range(len(hydraulicLines)):
            if "Black box component temperatures" in hydraulicLines[i]:
                j = i + 1
                while hydraulicLines[j] != "\n":
                    blackBoxLines.append(hydraulicLines[j])
                    j += 1
                break
        messageText = "Is this correct?\n\n"
        for line in blackBoxLines:
            messageText += line
        qmb = QMessageBox()
        qmb.setText(messageText)
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        qmb.setDefaultButton(QMessageBox.No)
        ret = qmb.exec()
        if ret == QMessageBox.No:
            qmb = QMessageBox()
            qmb.setText(
                "Please make sure the black box component temperatures are correct in hydraulic.ddck before starting a simluation."
            )
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return

        #   Check ddcks of storage tanks
        storageWithoutFile = []
        for object in self.centralWidget.trnsysObj:
            if isinstance(object, StorageTank):
                storageTankFile = os.path.join(
                    object.displayName, object.displayName + ".ddck"
                )
                storageTankPath = os.path.join(ddckPath, storageTankFile)
                if not (os.path.isfile(storageTankPath)):
                    storageWithoutFile.append(object.displayName + "\n")

        if storageWithoutFile:
            messageText = "The following storage tank(s) do(es) not have a corresponding ddck:\n\n"
            for storage in storageWithoutFile:
                messageText += storage
            messageText += "\nPlease make sure you that you export the ddck for every storage tank before starting a simulation."
            qmb = QMessageBox()
            qmb.setText(messageText)
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return

        #   Update run.config
        self.updateRun()

        #   Start simulation
        runApp = RunMain()
        executionFailed, errorStatement = runApp.runAction(
            self.logger, self.centralWidget.projectFolder
        )

        if executionFailed:
            messageText = (
                "Exception while trying to execute RunParallelTrnsys:\n\n"
                + errorStatement
            )
            qmb = QMessageBox()
            qmb.setText(messageText)
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()

        return

    def processSimulation(self):
        processPath = os.path.join(self.projectFolder, "process.config")
        if not os.path.isfile(processPath):
            messageText = "No such file:\n" + processPath
            qmb = QMessageBox()
            qmb.setText(messageText)
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return
        processApp = ProcessMain()
        executionFailed, errorStatement = processApp.processAction(
            self.logger, self.centralWidget.projectFolder
        )

        if executionFailed:
            messageText = (
                "Exception while trying to execute RunParallelTrnsys:\n\n"
                + errorStatement
            )
            qmb = QMessageBox()
            qmb.setText(messageText)
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()

        return

    def renameDia(self):
        self.logger.info("Renaming diagram...")
        self.centralWidget.showDiagramDlg()

    def deleteDia(self):
        qmb = QMessageBox()
        qmb.setText(
            'Are you sure you want to delete the diagram? (There is no possibility to "undo".)'
        )
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()
        if ret == QMessageBox.Yes:
            self.logger.info("Deleting diagram")
            self.centralWidget.delBlocks()
        else:
            self.logger.info("Canceling")
            return

    def createGroup(self):
        # print("Tb createGroup pressed")
        # global selectionMode
        self.centralWidget.selectionMode = True
        self.centralWidget.groupMode = True
        self.centralWidget.copyMode = False
        self.centralWidget.multipleSelectMode = False

    def tidyUp(self):
        self.logger.info("Tidying up...")
        self.centralWidget.cleanUpConnections()

    def setZoomIn(self):
        self.logger.info("Setting zoom in")
        self.centralWidget.diagramView.scale(1.2, 1.2)

    def setZoomOut(self):
        self.logger.info("Setting zoom out")
        self.centralWidget.diagramView.scale(0.8, 0.8)

    def setZoom0(self):
        self.logger.info("Setting zoom 0")
        self.centralWidget.diagramView.resetTransform()

    def copySelection(self):
        self.logger.info("Copying selection")
        # global selectionMode
        # global copyMode

        self.centralWidget.selectionMode = True
        self.centralWidget.copyMode = True
        self.centralWidget.groupMode = False
        self.centralWidget.multipleSelectMode = False

        self.centralWidget.copyElements()

    def pasteSelection(self):
        self.logger.info("Pasting selection")
        self.centralWidget.pasteFromClipBoard()
        # global copyMode
        self.centralWidget.copyMode = False

    def editGroups(self):
        self.centralWidget.editGroups()

    def mb_debug(self):
        pass
        # print(self.centralWidget.trnsysObj)
        # a = [int(s.id) for s in self.centralWidget.trnsysObj]
        # a.sort()
        # print(a)
        #
        # for s in self.centralWidget.trnsysObj:
        #     if s.id == 11:
        #         print("Duplicate id obj is " + str(s.displayName) + " " + str(s.id))
        #
        # for t in self.centralWidget.trnsysObj:
        #     t.alignMode = True
        #     print(t.alignMode)

        # temp = []
        # for t in self.centralWidget.trnsysObj:
        #     if isinstance(t, BlockItem):
        #         for p in t.inputs + t.outputs:
        #             if p not in temp:
        #                 temp.append(p)
        #
        # for p in temp:
        #     if p.isFromHx:
        #         print("Port with parent " + str(p.parent.displayName) + "is from Hx")
        #
        # res = True
        #
        # for b in self.centralWidget.trnsysObj:
        #     if isinstance(b, Connection) and b not in self.centralWidget.connectionList:
        #         res = False
        #
        # print("editor connectionList is consistent with trnsysObj: " + str(res))

        # dIns = DeepInspector(self.centralWidget)

    def runAndVisMf(self):
        self.calledByVisualizeMf = True
        filePaths = self.runMassflowSolver()
        if not filePaths:
            self.logger.error("Could not execute runMassflowSolver")
            return

        mfrFile, tempFile = filePaths
        if not os.path.isfile(mfrFile) or not os.path.isfile(tempFile):
            self.logger.error("No mfrFile or tempFile found!")
            return

        MassFlowVisualizer(self, mfrFile, tempFile)
        self.massFlowEnabled = True

    def visualizeMf(self):
        qmb = QMessageBox()
        qmb.setText(
            "Please select the mass flow rate prt-file that you want to visualize."
        )
        qmb.setStandardButtons(QMessageBox.Ok)
        qmb.setDefaultButton(QMessageBox.Ok)
        qmb.exec()

        mfrFile = QFileDialog.getOpenFileName(self, "Open diagram", filter="*Mfr.prt")[
            0
        ].replace("/", "\\")
        tempFile = mfrFile.replace("Mfr", "T")
        self.calledByVisualizeMf = True
        if os.path.isfile(mfrFile) and os.path.isfile(tempFile):
            MassFlowVisualizer(self, mfrFile, tempFile)
            self.massFlowEnabled = True
        else:
            self.logger.info("No mfrFile or tempFile found!")

    def openFile(self):
        self.logger.info("Opening diagram")
        qmb = QMessageBox()
        qmb.setText(
            "Are you sure you want to open another project? Unsaved progress on the current one will be lost."
        )
        qmb.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()

        if ret == QMessageBox.Cancel:
            return

        maybeCancelled = _prj.getLoadOrMigrateProject()
        if _ccl.isCancelled(maybeCancelled):
            return
        project = _ccl.value(maybeCancelled)

        self.centralWidget = self._createDiagramEditor(project)
        self.setCentralWidget(self.centralWidget)

    def toggleConnLabels(self):
        self.labelVisState = not self.labelVisState
        self.centralWidget.setConnLabelVis(self.labelVisState)

    def exportHydraulicsDdck(self):
        statusQuo = self.labelVisState
        if not statusQuo:
            self.toggleConnLabels()
        self.centralWidget.exportHydraulics(exportTo="ddck")
        if not statusQuo:
            self.toggleConnLabels()

    def exportHydraulicControl(self):
        self.centralWidget.exportHydraulicControl()

    def exportDck(self):
        dckBuilder.buildDck(self.projectFolder)

    def toggleEditorMode(self):
        self.logger.info("Toggling editor mode")
        self.centralWidget.editorMode = (self.centralWidget.editorMode + 1) % 2
        self.sb.showMessage("Mode is " + str(self.centralWidget.editorMode))

    def toggleAlignMode(self):
        self.logger.info("Toggling alignMode")
        self.centralWidget.alignMode = not self.centralWidget.alignMode

    def toggleSnap(self):
        self.centralWidget.snapGrid = not self.centralWidget.snapGrid
        self.centralWidget.diagramScene.update()

    def createSelection(self):
        self.centralWidget.clearSelectionGroup()
        self.centralWidget.selectionMode = True
        self.centralWidget.copyMode = False
        self.centralWidget.groupMode = False
        self.centralWidget.multipleSelectMode = True

    def deleteMultiple(self):
        # print("pressed del")
        temp = []
        self.logger.info("Child Items")
        self.logger.info(self.centralWidget.selectionGroupList.childItems())

        for t in self.centralWidget.selectionGroupList.childItems():
            temp.append(t)
            self.centralWidget.selectionGroupList.removeFromGroup(t)

        for t in temp:
            if isinstance(t, BlockItem):
                t.deleteBlock()
            elif isinstance(t, Connection):
                t.deleteConn()
            elif isinstance(t, GraphicalItem):
                t.deleteBlock()
            else:
                self.logger.info("Neiter a Block nor Connection in copyGroupList ")

    def runMassflowSolver(self):
        self.logger.info("Running massflow solver...")

        exportPath = self.centralWidget.exportHydraulics(exportTo="mfs")
        self.exportedTo = exportPath
        self.logger.info(exportPath)
        if exportPath != "None":
            msgb = QMessageBox(self)
            if not self.centralWidget.trnsysPath.is_file():
                msgb.setText(
                    "TRNExe.exe not found! Consider correcting the path in the settings."
                )
                msgb.exec()
                return 0, 0
            self.logger.info("trnsyspath: %s", self.centralWidget.trnsysPath)
            # cmd = f"{self.centralWidget.trnsysPath} {exportPath} /H"
            errorStatement = ""
            try:
                subprocess.run(
                    [str(self.centralWidget.trnsysPath), exportPath, "/H"], check=True
                )
                mfrFile = os.path.join(
                    self.projectFolder, self.projectFolder.split("\\")[-1] + "_Mfr.prt"
                )
                tempFile = os.path.join(
                    self.projectFolder, self.projectFolder.split("\\")[-1] + "_T.prt"
                )
                self.calledByVisualizeMf = False
                return mfrFile, tempFile
            except ValueError as e:
                self.logger.error("EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys")
                for words in e.args:
                    errorStatement += str(words)
            except OSError as e:
                self.logger.error("EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys")
                errorStatement = str(e)
            except:
                self.logger.error(
                    "UNDEFINED EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys"
                )

            messageText = (
                "Exception while trying to execute runMassflowSolver:\n\n"
                + errorStatement
            )
            qmb = QMessageBox()
            qmb.setText(messageText)
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()

            return None

    def loadVisualization(self):
        MfrFile = QFileDialog.getOpenFileName(
            self, "Select Mfr File", "exports", filter="*_Mfr.prt"
        )[0]
        if MfrFile == "":
            msgb = QMessageBox(self)
            msgb.setText("No Mfr file chosen!")
            msgb.exec()
            return
        TempFile = QFileDialog.getOpenFileName(
            self, "Select Temperature File", "exports", filter="*_T.prt"
        )[0]
        if TempFile == "":
            msgb = QMessageBox(self)
            msgb.setText("No Temperature file chosen!")
            msgb.exec()
            return

        selectedMfrFileName = str(MfrFile).split("/")[-1][:-8]
        selectedTempFileName = str(TempFile).split("/")[-1][:-6]

        currentFilePath = self.currentFile
        if "\\" in currentFilePath:
            diaName = currentFilePath.split("\\")[-1][:-5]
        elif "/" in currentFilePath:
            diaName = currentFilePath.split("/")[-1][:-5]
        else:
            diaName = currentFilePath

        if selectedMfrFileName == selectedTempFileName == diaName:
            MassFlowVisualizer(self, MfrFile, TempFile)
            self.massFlowEnabled = True
        else:
            self.logger.info(selectedMfrFileName, selectedTempFileName, diaName)
            msgb = QMessageBox(self)
            msgb.setText(
                "MFR or Temperature file does not correspond to current diagram!"
            )
            msgb.exec()

    def movePorts(self):
        self.centralWidget.moveDirectPorts = True

    def mouseMoveEvent(self, e):
        pass
        # x = e.x()
        # y = e.y()
        #
        # text = "x: {0},  y: {1}".format(x, y)
        # self.sb.showMessage(text)
        # #print("event")

    def showAbout(self):
        msgb = QMessageBox(self)
        msgb.setText("PyQt based diagram editor coupled to Trnsys functions")
        msgb.exec()

    def showVersion(self):
        msgb = QMessageBox(self)
        msgb.setText(
            "Currrent version is " + __version__ + " with status " + __status__
        )
        msgb.exec()

    def showCredits(self):
        msgb = QMessageBox(self)
        msgb.setText(
            "<p><b>Contributors:</b></p>"
            "<p>Stefano Marti, Dani Carbonell, Mattia Battaglia, Jeremias Schmidli, and Martin Neugebauer."
        )
        # "Icons made by Jeremias Schmidli and with icons by Vaadin from  www.flaticon.com</p>"
        msgb.exec()

    def exportPDF(self):
        self.centralWidget.printPDF()

    def closeEvent(self, e):
        qmb = QMessageBox()
        qmb.setText(
            "Do you want to save the current state of the project before closing the program?"
        )
        qmb.setStandardButtons(
            QMessageBox.Save | QMessageBox.Close | QMessageBox.Cancel
        )
        qmb.setDefaultButton(QMessageBox.Cancel)
        ret = qmb.exec()
        if ret == QMessageBox.Cancel:
            e.ignore()
        elif ret == QMessageBox.Close:
            e.accept()
        elif ret == QMessageBox.Save:
            self.centralWidget.save()
            e.accept

    def ensureSettingsExist(self):
        settings = _settings.Settings.tryLoadOrNone()
        if not settings:
            self.askUserForSettingsValuesAndSave()

    def askUserForSettingsValuesAndSave(self):
        newSettings = _sdlg.SettingsDlg.showDialogAndGetSettings(parent=self)
        while newSettings == _sdlg.CANCELLED:
            newSettings = _sdlg.SettingsDlg.showDialogAndGetSettings(parent=self)
        newSettings.save()

    def loadTrnsysPath(self):
        settings = _settings.Settings.load()
        self.centralWidget.trnsysPath = _pl.Path(settings.trnsysBinaryPath)

    def debugConns(self):
        """
        Check each block items for error connections.
        Returns warning message if blockitem contains two input connections or two output connections
        """
        self.logger.info("trnsysObjs:", self.centralWidget.trnsysObj)
        self.noErrorConns = True
        for o in self.centralWidget.trnsysObj:
            self.logger.info(o)
            if isinstance(o, BlockItem) and len(o.outputs) == 1 and len(o.inputs) == 1:
                self.logger.info("Checking block connections", o.displayName)
                objInput = o.inputs[0]
                objOutput = o.outputs[0]
                connToInputToPort = objInput.connectionList[0].toPort
                connToOutputToPort = objOutput.connectionList[0].toPort
                connToInputFromPort = objInput.connectionList[0].fromPort
                connToOutputFromPort = objOutput.connectionList[0].fromPort
                connName1 = objInput.connectionList[0].displayName
                connName2 = objOutput.connectionList[0].displayName
                objName = o.displayName

                if objInput == connToInputToPort and objOutput == connToOutputToPort:
                    msgBox = QMessageBox()
                    msgBox.setText(
                        "both %s and %s are input ports into %s"
                        % (connName1, connName2, objName)
                    )
                    msgBox.exec_()
                    self.noErrorConns = False

                elif (
                    objInput == connToInputFromPort
                    and objOutput == connToOutputFromPort
                ):
                    msgBox = QMessageBox()
                    msgBox.setText(
                        "both %s and %s are output ports from %s"
                        % (connName1, connName2, objName)
                    )
                    msgBox.exec_()
                    self.noErrorConns = False
        return self.noErrorConns

    def runApp(self):
        runApp = RunMain()
        if self.centralWidget.projectPath == "":
            self.logger.info("Temp path:", self.centralWidget.projectFolder)
            runApp.runAction(self.centralWidget.projectFolder)
        else:
            self.logger.info("Project path:", self.centralWidget.projectPath)
            runApp.runAction(self.centralWidget.projectPath)

    def _createDiagramEditor(self, project: _prj.Project) -> _de.Editor:
        if isinstance(project, _prj.LoadProject):
            self.projectFolder = str(project.jsonFilePath.parent)
            return _de.Editor(
                self,
                str(project.jsonFilePath.parent),
                jsonPath=None,
                loadValue="load",
                logger=self.logger,
            )

        if isinstance(project, _prj.MigrateProject):
            self.projectFolder = str(project.newProjectFolderPath)
            editor = _de.Editor(
                self,
                str(project.newProjectFolderPath),
                str(project.oldJsonFilePath),
                loadValue="json",
                logger=self.logger,
            )
            editor.save()
            return editor

        if isinstance(project, _prj.CreateProject):
            self.projectFolder = str(project.jsonFilePath.parent)
            return _de.Editor(
                self,
                self.projectFolder,
                jsonPath=str(project.jsonFilePath),
                loadValue="new",
                logger=self.logger,
            )

        raise AssertionError(f"Unknown `project' type: {type(project)}")


def main():
    arguments = args.getArgsOrExit()

    logger = log.setup_custom_logger("root", arguments.logLevel)
    app = QApplication(sys.argv)
    app.setApplicationName("Diagram Creator")

    maybeCancelled = _prj.getProject()
    if _ccl.isCancelled(maybeCancelled):
        return
    project = _ccl.value(maybeCancelled)

    form = _MainWindow(logger, project)
    form.showMaximized()
    form.show()
    form.ensureSettingsExist()
    form.loadTrnsysPath()

    tracer = trc.createTracer(arguments.shallTrace)
    tracer.run(lambda: app.exec())


if __name__ == "__main__":
    main()



# pylint: skip-file
# type: ignore

__all__ = ["SettingsDlg", "MaybeCancelled", "CANCELLED"]

import pathlib as _pl
import typing as _tp

import PyQt5.QtCore as _qtc
import PyQt5.QtWidgets as _widgets

import trnsysGUI.settings as _settings


class _Cancelled:
    pass


CANCELLED = _Cancelled()

_TCo = _tp.TypeVar("_TCo", covariant=True)

MaybeCancelled = _tp.Union[_TCo, _Cancelled]


class SettingsDlg(_widgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        settings = self._getSettings()
        oldTrnsysPath = settings.trnsysBinaryPath if settings else ""

        label = _widgets.QLabel("Trnsys Path:")
        self._lineEdit = _widgets.QLineEdit(oldTrnsysPath)
        self._lineEdit.setDisabled(True)

        setButton = _widgets.QPushButton("Set Trnsys Path")
        setButton.setFixedWidth(100)
        setButton.clicked.connect(self._onSetButtonClicked)

        layout = _widgets.QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self._lineEdit)
        layout.addWidget(setButton)

        self._okButton = _widgets.QPushButton("Done")
        self._okButton.setFixedWidth(50)
        self._okButton.clicked.connect(self._onOkButtonClicked)
        canOk = bool(settings)
        self._okButton.setEnabled(canOk)
        buttonLayout = _widgets.QHBoxLayout()
        buttonLayout.addWidget(self._okButton, alignment=_qtc.Qt.AlignCenter)

        overallLayout = _widgets.QVBoxLayout()
        overallLayout.addLayout(layout)
        overallLayout.addLayout(buttonLayout)
        self.setLayout(overallLayout)
        self.setFixedWidth(800)

        self.setWindowTitle("Set Paths")

        self._settings: MaybeCancelled[_settings.Settings] = CANCELLED

    @staticmethod
    def showDialogAndGetSettings(parent=None) -> MaybeCancelled[_settings.Settings]:
        dialog = SettingsDlg(parent)
        dialog.exec()
        return dialog._settings

    @staticmethod
    def _getSettings() -> _tp.Optional[_settings.Settings]:
        return _settings.Settings.tryLoadOrNone()

    def _onSetButtonClicked(self) -> None:
        oldDirPath = self._lineEdit.text()

        newBinaryPath, _ = _widgets.QFileDialog.getOpenFileName(
            self, "Select TRNSYS exe", directory=oldDirPath, filter="*.exe"
        )

        self._lineEdit.setText(newBinaryPath)

        canOk = True if self._getSettings() or newBinaryPath else False
        self._okButton.setEnabled(canOk)

    def _onOkButtonClicked(self) -> None:
        newBinaryPath = self._lineEdit.text()

        settings = self._getSettings()
        if settings:
            settings.trnsysBinaryPath = newBinaryPath
        else:
            settings = _settings.Settings.create(_pl.Path(newBinaryPath))

        self._settings = settings

        self.close()

    def cancel(self) -> None:
        self._settings = CANCELLED
        self.close()



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QUndoCommand

from trnsysGUI.Connection import Connection


class CreateConnectionCommand(QUndoCommand):
    def __init__(self, fromPort, toPort, connParent, descr):
        super().__init__(descr)
        self.conn = None
        self.connFromPort = fromPort
        self.connToPort = toPort
        self.connParent = connParent

    def redo(self):
        self.conn = Connection(self.connFromPort, self.connToPort, self.connParent)

    def undo(self):
        if self.conn in self.conn.parent.connectionList:
            self.conn.deleteConn()



# pylint: skip-file
# type: ignore

from PyQt5 import QtCore
from PyQt5.QtCore import QRectF, QSize, QPointF, Qt
from PyQt5.QtGui import QBrush, QColor, QPen
from PyQt5.QtWidgets import QGraphicsEllipseItem


class ResizerItem(QGraphicsEllipseItem):
    def __init__(self, parent):
        QGraphicsEllipseItem.__init__(self, QRectF(-4, -4, 8, 8), parent)
        self.parent = parent
        self.setPos(self.parent.w, self.parent.h)
        self.setBrush(Qt.transparent)
        p1 = QPen()
        p1.setStyle(Qt.DotLine)
        self.setPen(p1)
        self.setFlags(self.ItemIsSelectable | self.ItemIsMovable)
        self.setFlag(self.ItemSendsScenePositionChanges, True)

    def itemChange(self, change, value):
        if change == self.ItemPositionChange:
            # print("relative pos has changed")
            self.parent.setItemSize(self.pos().x(), self.pos().y())
            self.parent.updateImage()
            return QPointF(value.x(), value.x())

        return super(ResizerItem, self).itemChange(change, value)

    # def mouseReleaseEvent(self, event):
    #     self.setPos(self.parent.boundingRect().bottomRight())
    #     super(ResizerItem, self).mouseReleaseEvent(event)
    #
    # def dragMoveEvent(self, event):
    #     self.setPos(event.pos().x(), event.pos().x())
    #     print("sdfhs")
    #     super(ResizerItem, self).dragMoveEvent(event)
    def delete(self):
        del self



# pylint: skip-file
# type: ignore

from PyQt5 import QtCore
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QGraphicsEllipseItem

from trnsysGUI.Node import Node


class CornerItem(QGraphicsEllipseItem):
    def __init__(self, x, y, r1, r2, prevNode, nextNode, parent=None):
        """
        CornerItems represent corners for each Connection.
        When a segmentItem is dragged, it will also move the corners, which will trigger the CornerItem
        itemChange method of Qt. There, the adjacent segments are redrawn.

        Note: Using hasattr to detect type of parent of node, better solution needed
        Note: No support for disr (interrupted) segements

        Parameters
        ----------
        x
        y
        r1 : int
        semi-axis length in x direction
        r2 : int
        semi-axis length in y direction
        prevNode
        nextNode
        parent
        """
        super(CornerItem, self).__init__(x, y, r1, r2, None)

        self.logger = parent.logger

        self.logger.debug("init pos is " + str(self.pos()))
        self.parent = parent
        self.setBrush(QBrush(QtCore.Qt.black))
        self.node = Node(self, prevNode, nextNode)

        self.posCallbacks = []

    def itemChange(self, change, value):
        # self.logger.debug("Change is " + str(change))
        if change == self.ItemScenePositionHasChanged:
            # self.logger.debug("pos change")
            # self.logger.debug("self has partial lenght " + str(self.parent.partialLength(self)))
            # self.logger.debug("Self is at " + str(self.parent.partialLength(self.node) / (self.parent.totalLength())) + "% of the whole connection")

            nNode = self.node.nextN()
            pNode = self.node.prevN()

            positionInArr = self.parent.getNodePos(self)

            segBefore = self.parent.segments[positionInArr]
            segAfter = self.parent.segments[positionInArr + 1]

            # self.logger.debug("Position in arr is" + str(positionInArr))

            if type(nNode.parent) is CornerItem:
                segAfter.setLine(
                    self.scenePos().x(), self.scenePos().y(), segAfter.line().p2().x(), segAfter.line().p2().y()
                )

            if type(pNode.parent) is CornerItem:
                segBefore.setLine(
                    segBefore.line().p1().x(), segBefore.line().p1().y(), self.scenePos().x(), self.scenePos().y()
                )

            if hasattr(nNode.parent, "fromPort"):
                # self.logger.debug("The moving node (cornerItem) is " + str(self.node))
                # self.logger.debug("type of nNode is conn")

                # We have two options: either nNode is at a port or at a disrupted segment
                if nNode.nextN() is not None:
                    # We are at a disrupted segment
                    pass
                    # segments = nNode.parent.segments
                    #
                    # s = None
                    # s2 = None
                    #
                    # for x in segments:
                    #     if x.endNode is nNode:
                    #         s = x
                    #     if x.startNode is nNode.nextN():
                    #         s2 = x
                    #
                    # if s is None or s2 is None:
                    #     self.logger.debug("Error, there shouldn't be a empty s or s2")
                    #
                    # s.startNode.setNext(s2.endNode)
                    # s2.endNode.setPrev(s.startNode)
                    #
                    # newS = segmentItem(s.startNode, s2.endNode, s.parent)
                    # newS.setVisible(False)
                    # s.parent.parent.diagramScene.addItem(newS)
                    #
                    # if type(s2.endNode.parent) is Connection:
                    #     self.logger.debug("Line of s is " + str(s.line()))
                    #     self.logger.debug("Line of s2 is " + str(s2.line()))
                    #     newS.setLine(self.scenePos().x(),
                    #                  self.scenePos().y(),
                    #                  s2.line().p2().x(),
                    #                  s2.line().p2().y())
                    #     self.logger.debug("Set pos node to Connection")
                    #
                    # elif type(s2.endNode.parent) is CornerItem:
                    #     newS.setLine(self.scenePos().x(),
                    #                  self.scenePos().y(),
                    #                  s2.line().p2().x(),
                    #                  s2.line().p2().y())
                    #     self.logger.debug("Set pos node to node")
                    #
                    # else:
                    #     pass
                    #
                    # newS.setVisible(True)
                    # s.hide()
                    # s2.hide()
                    # s.parent.segments.remove(s2)
                    # s.parent.segments.remove(s)
                    # # self.logger.debug("Length of segments is " + str(len(segments)))
                    #
                    # del s2.endNode
                    # del s.endNode
                    #
                    # newS.parent.parent.diagramScene.removeItem(s)
                    # newS.parent.parent.diagramScene.removeItem(s2)
                    #
                    # newS.parent.buildBridges()

                else:

                    if self.node.lastNode() is nNode:
                        self.logger.debug("nNode is at toPort")
                        t = self.parent.segments[-1]
                        t.setLine(
                            self.scenePos().x(),
                            self.scenePos().y(),
                            nNode.parent.toPort.scenePos().x(),
                            nNode.parent.toPort.scenePos().y(),
                        )

            if hasattr(pNode.parent, "fromPort"):
                # self.logger.debug("The moving node (cornerItem) is " + str(self.node))
                # self.logger.debug("type of pNode is conn")

                if pNode.prevN() is not None:
                    # We are at a disrupted segment
                    pass
                    # segments = nNode.parent.segments
                    #
                    # s = None
                    # s2 = None
                    #
                    # for x in segments:
                    #     if x.startNode is pNode:
                    #         s2 = x
                    #     if x.endNode is pNode.prevN():
                    #         s = x
                    #
                    # if s is None or s2 is None:
                    #     self.logger.debug("Error, there shouldn't be a empty s or s2")
                    #
                    # s.startNode.setNext(s2.endNode)
                    # s2.endNode.setPrev(s.startNode)
                    #
                    # newS = segmentItem(s.startNode, s2.endNode, s.parent)
                    # newS.setVisible(False)
                    # s.parent.parent.diagramScene.addItem(newS)
                    #
                    # if type(s2.endNode.parent) is Connection:
                    #     self.logger.debug("Line of s is " + str(s.line()))
                    #     self.logger.debug("Line of s2 is " + str(s2.line()))
                    #     newS.setLine(s.line().p2().x(),
                    #                  s.line().p2().y(), self.scenePos().x(), self.scenePos().y())
                    #
                    #     self.logger.debug("Set pos node to Connection")
                    #
                    # elif type(s2.endNode.parent) is CornerItem:
                    #     newS.setLine(s.line().p1().x(),
                    #                  s.line().p1().y(), self.scenePos().x(), self.scenePos().y())
                    #     self.logger.debug("Set pos node to node")
                    #
                    # else:
                    #     pass
                    #
                    # newS.setVisible(True)
                    # s.hide()
                    # s2.hide()
                    # s.parent.segments.remove(s2)
                    # s.parent.segments.remove(s)
                    # # self.logger.debug("Length of segments is " + str(len(segments)))
                    #
                    # del s2.endNode
                    # del s.endNode
                    #
                    # newS.parent.parent.diagramScene.removeItem(s)
                    # newS.parent.parent.diagramScene.removeItem(s2)
                    #
                    # newS.parent.buildBridges()

                else:
                    if self.node.firstNode() is pNode:
                        self.logger.debug("pNode is at fromPort")
                        f = self.parent.segments[0]
                        f.setLine(
                            pNode.parent.fromPort.scenePos().x(),
                            pNode.parent.fromPort.scenePos().y(),
                            self.scenePos().x(),
                            self.scenePos().y(),
                        )

            for s in self.parent.segments:
                s.updateGrad()

        if change == self.ItemScenePositionHasChanged:
            for cb in self.posCallbacks:
                cb(value)
            return value
        return super(CornerItem, self).itemChange(change, value)



# pylint: skip-file
# type: ignore

from PyQt5.QtCore import QPointF
from PyQt5.QtWidgets import QGraphicsRectItem, QGraphicsTextItem

from trnsysGUI.Connection import Connection
from trnsysGUI.PortItem import PortItem
from trnsysGUI.segmentItem import segmentItem
from trnsysGUI.BlockItem import BlockItem


class Group(QGraphicsRectItem):
    def __init__(self, x, y, w, h, parent):
        super(Group, self).__init__(x, y, w, h)
        self.parent = parent
        self.x = x
        self.y = y
        self.w = w
        self.h = h

        self.displayName = "Untitled"

        # List of items part of this group
        self.itemList = []

        self.exportDi = 0.05  # [m]
        self.exportL = 2  # [m]
        self.exportU = 10  # DC will calculate proper default value

        self.label = QGraphicsTextItem(self)
        self.label.setPlainText(self.displayName)

        self.setVisible(False)

        # Add rect to scene
        self.parent.addItem(self)

        # Add group to groupList
        self.parent.parent().groupList.append(self)

        # Update the label position
        self.updateLabelPos()

    def computeRect(self):
        factor = 1.2
        center = None

        blockitems = 0

        max_x = 0
        max_y = 0
        min_x = 0
        min_y = 0

        for i in self.itemList:
            if isinstance(i, BlockItem):
                center += i.scenePos()
                blockitems += 1
                if i.scenePos().x() > max_x:
                    max_x = i.scenePos().x()

                if i.scenePos().x() < min_x:
                    min_x = i.scenePos().x()

                if i.scenePos().y() > max_y:
                    max_y = i.scenePos().y()

                if i.scenePos().x() < min_x:
                    min_y = i.scenePos().y()

            if type(i) is Connection:
                # ignore for center calculation
                pass

        center = QPointF(1 / blockitems * center.x(), 1 / blockitems * center.y())

        return (
            center.x() - (max_x - min_x) / 2 * factor,
            center.y() - (max_y - min_y) / 2 * factor,
            (max_x - min_x) * factor,
            (max_y - min_y) * factor,
        )

    def setItemsGroup(self, itemList):
        """
        Sets the items to this group
        Parameters
        ----------
        itemList

        Returns
        -------

        """
        for o in itemList:
            if isinstance(o, BlockItem):
                o.setBlockToGroup(self.displayName)
            elif isinstance(o, Connection):
                o.setConnToGroup(self.displayName)
            else:
                print("Found an item which is wether Block nor Connection in setItemsGroup")

    def addBlock(self, bl):
        inGroup = False

        for i in self.itemList:
            if isinstance(i, BlockItem):
                if i.id == bl.id:
                    print("Found the block already to be in group")
                    inGroup = True

        if not inGroup:
            print("Block not yet in this group")
            bl.append(bl)
            self.setItemsGroup()

    def addConnection(self, c):
        inGroup = False

        for i in self.itemList:
            if isinstance(i, Connection):
                if i.id == c.id:
                    print("Found the connection already to be in group")
                    inGroup = True

        if not inGroup:
            print("Block not yet in this group")
            c.append(c)
            self.setItemsGroup()

    def setName(self, newName):
        self.displayName = newName
        self.label.setPlainText(newName)
        # self.setItemsGroup()

    def updateLabelPos(self):
        factor = 0.9
        self.label.setPos(self.x + factor * self.w, self.y + self.h + 5)

    def deleteGroup(self):
        self.parent.parent().groupList.remove(self)
        self.parent.removeItem(self)

    def printGroup(self):
        print("Printout of group " + str(self) + str(self.displayName))
        print(
            "There are "
            + len([t for t in self.parent.parent().trnsysObj if type(t) is BlockItem])
            + " blocks and "
            + len([t for t in self.parent.parent().trnsysObj if type(t) is Connection])
            + " connections"
        )

        print("The elements are " + [t.displayName for t in self.parent.parent().trnsysObj])



# pylint: skip-file
# type: ignore

from itertools import count


class IdGenerator(object):
    # This value is used for the IDs when migrating from old JSON formats.
    # After deserialization the IDs are set such that there are no duplicate
    # IDs across all objects.
    UNINITIALIZED_ID = -1

    def __init__(self):
        self.ids = []
        self.trnsysIds = []
        self.connIds = []
        self.storageTes = []
        self.storageType = []
        self.idCounter = count(1)
        self.trnsysIdCounter = count(1)
        self.connIdCounter = count(1)
        self.storagenTesCounter = count(1)
        self.storageTypeCounter = count(1924)

    def getID(self):
        id = next(self.idCounter)
        self.ids.append(id)
        return id

    def getTrnsysID(self):
        id = next(self.trnsysIdCounter)
        self.trnsysIds.append(id)
        return id

    def getConnID(self):
        id = next(self.connIdCounter)
        self.connIds.append(id)
        return id

    def getStoragenTes(self):
        id = next(self.storagenTesCounter)
        self.storageTes.append(id)
        return id

    def getStorageType(self):
        id = next(self.storageTypeCounter)
        self.storageType.append(id)
        return id

    def setID(self, id):
        self.idCounter = count(id)

    def setConnID(self, id):
        self.connIdCounter = count(id)

    def setTrnsysID(self, id):
        self.trnsysIdCounter = count(id)

    def reset(self):
        self.ids = []
        self.trnsysIds = []
        self.connIds = []
        self.storageTes = []
        self.storageType = []
        self.idCounter = count(1)
        self.trnsysIdCounter = count(1)
        self.connIdCounter = count(1)
        self.storagenTesCounter = count(1)
        self.storageTypeCounter = count(1924)



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import (
    QLabel,
    QLineEdit,
    QPushButton,
    QHBoxLayout,
    QGridLayout,
    QDialog,
    QRadioButton,
    QVBoxLayout,
    QListWidget,
)


class GenericPortPairDlg(QDialog):
    def __init__(self, block, parent=None):
        super(GenericPortPairDlg, self).__init__(parent)
        self.block = block
        nameLabel = QLabel("Name:")
        objectLabel = QLabel("Object:" + str(self.block))
        self.le = QLineEdit("0")

        positionLayout = QVBoxLayout()
        positionLabel = QLabel("Select a position")

        self.rButtonLeft = QRadioButton("Left")
        self.rButtonTop = QRadioButton("Top")
        self.rButtonRight = QRadioButton("Right")
        self.rButtonBottom = QRadioButton("Bottom")

        positionLayout.addWidget(positionLabel)
        positionLayout.addWidget(self.rButtonLeft)
        positionLayout.addWidget(self.rButtonTop)
        positionLayout.addWidget(self.rButtonRight)
        positionLayout.addWidget(self.rButtonBottom)

        self.addButton = QPushButton("Add port pair")
        self.cancelButton = QPushButton("Close")

        self.addButton.clicked.connect(self.addPortPair)
        self.cancelButton.clicked.connect(self.cancel)

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.addButton)
        buttonLayout.addWidget(self.cancelButton)

        radioButtonLayout = QHBoxLayout()
        self.oButton = QRadioButton("Output")
        self.iButton = QRadioButton("Input")
        radioButtonLayout.addWidget(self.iButton)
        radioButtonLayout.addWidget(self.oButton)

        listWLabel = QLabel("Ports pairs")
        self.listW = QListWidget()

        layout = QGridLayout()
        # layout.addWidget(nameLabel, 0, 0)
        # layout.addWidget(self.le, 0, 1)
        # layout.addLayout(radioButtonLayout, 1, 0, 2, 0)
        layout.addWidget(listWLabel, 0, 0)
        layout.addWidget(self.listW, 1, 0)
        layout.addLayout(positionLayout, 2, 0, 2, 0)
        layout.addLayout(buttonLayout, 5, 0, 2, 0)
        self.setLayout(layout)

        self.setWindowTitle("Change connection name")
        self.loadList()
        self.show()

    def loadList(self):
        print("Loading ports")
        stringDct = {"0": "Left", "1": "Top", "2": "Right", "3": "Bottom"}
        for i in self.block.inputs:
            self.listW.addItem("Port pair at " + stringDct[str(i.side)])

    def addPortPair(self):
        if self.getSide() is not None:
            print("Adding a port")
            self.block.addPortPair(self.getSide())

    def removePortPair(self):
        if self.getSide() is not None:
            self.block.removePortPair(self.getSide())

    def getSide(self):
        if self.rButtonLeft.isChecked():
            return 0
        elif self.rButtonTop.isChecked():
            return 1
        elif self.rButtonRight.isChecked():
            return 2
        elif self.rButtonBottom.isChecked():
            return 3
        else:
            return None

    def cancel(self):
        self.close()



# pylint: skip-file
# type: ignore

import argparse as _ap
import dataclasses as _dc


__all__ = ["getArgsOrExit", "Arguments"]


def getArgsOrExit() -> "Arguments":
    logLevels = "CRITICAL ERROR WARNING INFO DEBUG".split()

    parser = _ap.ArgumentParser()
    parser.add_argument("-l", "--log", default="INFO", choices=logLevels, help="Set the log level", metavar="LEVEL")
    parser.add_argument("-t", "--trace", action="store_true", help="Enable tracing")

    namespace = parser.parse_args()
    logLevel = namespace.log
    shallTrace = namespace.trace

    return Arguments(logLevel, shallTrace)


@_dc.dataclass(frozen=True)
class Arguments:
    logLevel: str
    shallTrace: bool



__all__ = ["Error", "Result", "isError", "value"]

import dataclasses as _dc
import typing as _tp

import PyQt5.QtWidgets as _qtw

_T = _tp.TypeVar("_T")


@_dc.dataclass
class Error:
    message: str


Result = _tp.Union[_T, Error]


def isError(result: Result[_T]) -> bool:
    if isinstance(result, Error):
        return True

    return False


def value(result: Result[_T]) -> _T:
    if isError(result):
        _error = _tp.cast(Error, result)
        raise ValueError(f"Cannot get value of an error. Error is {_error.message}")

    return _tp.cast(_T, result)


def error(result: Result[_T]) -> Error:
    if not isError(result):
        raise ValueError("Result is not an error.")

    return _tp.cast(Error, result)


def showErrorMessageBox(_error: Error) -> None:
    messageBox = _qtw.QMessageBox()
    messageBox.setText(_error.message)
    messageBox.exec()



# pylint: skip-file
# type: ignore

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QDialog,
    QLabel,
    QLineEdit,
    QPushButton,
    QCheckBox,
    QHBoxLayout,
    QGridLayout,
    QTabWidget,
    QWidget,
    QDoubleSpinBox,
    QMessageBox,
)

import trnsysGUI.TVentil as _vnt


class TVentilDlg(QDialog):
    def __init__(self, ventil: _vnt.TVentil, parent=None):
        super(TVentilDlg, self).__init__(parent)
        nameLabel = QLabel("Name:")

        self.logger = parent.logger

        self.block = ventil
        self.valvePosition = self.block.positionForMassFlowSolver
        self.le = QLineEdit(self.block.label.toPlainText())
        self.setWindowIcon(QIcon(ventil.pixmap()))
        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        self.hFlipBoxLabel = QLabel("Horizontal Flip")
        self.vFlipBoxLabel = QLabel("Vertical Flip")
        self.complexDivLabel = QLabel("Tempering valve")
        self.hFlipBox = QCheckBox()
        self.vFlipBox = QCheckBox()
        self.complexDiv = QCheckBox()

        self.vFlipBox.setTristate(False)
        self.hFlipBox.setTristate(False)
        self.hFlipBox.setCheckState(self.block.flippedH * 2)
        self.vFlipBox.setCheckState(self.block.flippedV * 2)
        self.complexDiv.setTristate(False)
        self.logger.debug("complexdiv is " + str(self.block.isTempering))
        self.complexDiv.setCheckState(self.block.isTempering * 2)

        flipLayout = QHBoxLayout()
        flipLayout.addWidget(self.hFlipBoxLabel)
        flipLayout.addWidget(self.hFlipBox)
        flipLayout.addWidget(self.vFlipBoxLabel)
        flipLayout.addWidget(self.vFlipBox)
        flipLayout.addWidget(self.complexDivLabel)
        flipLayout.addWidget(self.complexDiv)

        textLayout = QHBoxLayout()
        self.warningLabel = QLabel("<b>Selecting tempering valve will always set Valve Position to 0.</b>")
        textLayout.addWidget(self.warningLabel)

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)
        tab1Layout = QGridLayout(self)
        tab1Layout.addWidget(nameLabel, 0, 0)
        tab1Layout.addWidget(self.le, 0, 1)
        tab1Layout.addLayout(flipLayout, 1, 0, 3, 0)
        tab1Layout.addLayout(textLayout, 3, 0, 3, 0)

        positionLayout = QHBoxLayout()
        self.positionMassFlowSolverLabel = QLabel("Valve Position")
        self.positionMassFlowSolver = QDoubleSpinBox()
        self.positionMassFlowSolver.setDecimals(1)
        self.positionMassFlowSolver.setSingleStep(0.1)
        self.positionMassFlowSolver.setProperty("value", self.valvePosition)
        self.positionMassFlowSolver.setRange(0, 1.0)
        self.positionMassFlowSolver.valueChanged.connect(self.positionMassFlowSolverChanged)
        positionLayout.addWidget(self.positionMassFlowSolverLabel)
        positionLayout.addWidget(self.positionMassFlowSolver)

        # Initialize tab screen
        self.tabs = QTabWidget()
        self.tab1 = QWidget()
        self.tab2 = QWidget()
        # Add tabs
        self.tabs.addTab(self.tab1, "Diagram")
        self.tabs.addTab(self.tab2, "Mass Flow Solver")
        self.tab1.setLayout(tab1Layout)
        self.tab2.setLayout(positionLayout)
        # self.tab2 = layout
        # self.tabs.resize(300, 200)

        # self.tabs.addTab(self.tab2, "Tab 2")
        self.layout2 = QGridLayout(self)
        self.layout2.addWidget(self.tabs, 0, 0)
        self.layout2.addLayout(buttonLayout, 2, 0, 3, 0)
        self.setLayout(self.layout2)

        self.setFixedSize(400, 180)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.hFlipBox.stateChanged.connect(self.setNewFlipStateH)
        self.vFlipBox.stateChanged.connect(self.setNewFlipStateV)
        self.complexDiv.stateChanged.connect(self.setNewComplexState)
        self.setWindowTitle("Properties")
        self.show()

    def acceptedEdit(self):
        self.logger.debug("Changing displayName")
        newName = self.le.text()
        if self.block.isTempering:
            self.block.setPositionForMassFlowSolver(0.0)
            self.block.posLabel.setPlainText(str(self.block.positionForMassFlowSolver))
        if newName.lower() == str(self.block.displayName).lower():
            self.block.posLabel.setPlainText(str(self.block.positionForMassFlowSolver))
            self.close()
        elif newName != "" and not self.nameExists(newName) and newName != self.block.displayName:
            # self.block.setName(newName)
            self.block.label.setPlainText(newName)
            self.block.displayName = newName
            self.block.posLabel.setPlainText(str(self.block.positionForMassFlowSolver))
            self.close()
        elif newName == "":
            msgb = QMessageBox()
            msgb.setText("Please Enter a name!")
            msgb.exec()
        elif self.nameExists(newName):
            msgb = QMessageBox()
            msgb.setText("Name already exist!")
            msgb.exec()

    def setNewFlipStateH(self, state):
        self.block.updateFlipStateH(state)

    def setNewFlipStateV(self, state):
        self.block.updateFlipStateV(state)

    def setNewComplexState(self, state):
        self.block.setComplexDiv(state)
        self.block.posLabel.setPlainText(str(self.block.positionForMassFlowSolver))

    def positionMassFlowSolverChanged(self, value):
        if self.block.isTempering:
            self.block.setPositionForMassFlowSolver(0.0)
        else:
            self.block.setPositionForMassFlowSolver(value)

    def cancel(self):
        self.close()

    def nameExists(self, n):
        for t in self.parent().trnsysObj:
            if str(t.displayName).lower() == n.lower():
                return True
        return False



# pylint: skip-file
# type: ignore

import typing as _tp
import math as _math

import numpy as np
from PyQt5.QtCore import QLineF, QPointF
from PyQt5.QtGui import QColor, QPen
from PyQt5.QtWidgets import QGraphicsTextItem, QUndoCommand

import trnsysGUI.BlockItem as _bi
import trnsysGUI.IdGenerator as _id
from trnsysGUI.Collector import Collector
from trnsysGUI.CornerItem import CornerItem
from trnsysGUI.Node import Node
from trnsysGUI.PortItem import PortItem
from trnsysGUI.Pump import Pump
from trnsysGUI.TVentil import TVentil
from trnsysGUI.segmentItem import segmentItem


def calcDist(p1, p2):
    vec = p1 - p2
    norm = _math.sqrt(vec.x() ** 2 + vec.y() ** 2)
    return norm


class Connection(object):
    def __init__(self, fromPort: PortItem, toPort: PortItem, parent, **kwargs):
        self.logger = parent.logger

        self.fromPort = fromPort
        self.toPort = toPort
        self.displayName = None

        self.parent = parent
        self.groupName = ""
        self.setDefaultGroup()

        # Export related
        self.typeNumber = 0
        self.exportConnsString = ""
        self.exportInputName = "0"
        self.exportInitialInput = -1
        self.exportEquations = []
        self._portItemsWithParent: _tp.List[_tp.Tuple[PortItem, _bi.BlockItem]] = []

        # Global
        self.id = self.parent.idGen.getID()
        self.connId = self.parent.idGen.getConnID()
        self.trnsysId = self.parent.idGen.getTrnsysID()

        self.segments: _tp.List[segmentItem] = []

        self.startNode = Node()
        self.endNode = Node()
        self.firstS: _tp.Optional[segmentItem] = None

        self.segmentsLoad = None
        self.cornersLoad = None
        self.labelPosLoad = None
        self.labelMassPosLoad = None

        self.mass = 0  # comment out
        self.temperature = 0

        # A new connection is created if there are no kwargs
        if not kwargs:
            self.fromPortId = self.parent.idGen.getID()
            self.toPortId = self.parent.idGen.getID()
            self.initNew(parent)
        else:
            if "loadedConn" in kwargs:
                self.logger.debug("Connection being loaded")
                self.fromPortId = kwargs["fromPortId"]
                self.toPortId = kwargs["toPortId"]
                self.segmentsLoad = kwargs["segmentsLoad"]
                self.cornersLoad = kwargs["cornersLoad"]

                if "labelPos" in kwargs:
                    self.labelPosLoad = kwargs["labelPos"]

                if "labelMassPos" in kwargs:
                    self.labelMassPosLoad = kwargs["labelMassPos"]

    def isVisible(self):
        res = True
        for s in self.segments:
            if not s.isVisible():
                res = False

        return res

    # Setter
    def setName(self, newName):
        self.displayName = newName
        for s in self.segments:
            s.label.setPlainText(newName)

    def setMassAndTemperature(self, mass, temp):
        """
        To show the mass and temperature during mass flow visualization
        """
        self.mass = float(mass)
        self.mass = "{:,}".format(self.mass)
        self.temperature = temp
        for s in self.segments:
            s.labelMass.setPlainText("M: %s kg/Hr   T: %s\u2103" % (self.mass.replace(",", "'"), self.temperature))

    def setDisplayName(self, newName):
        self.displayName = newName
        self.updateSegLabels()

    def setLabelPos(self, tup: _tp.Tuple[float, float]) -> None:
        pos = self._toPoint(tup)
        self.firstS.label.setPos(pos)

    def setMassLabelPos(self, tup: _tp.Tuple[float, float]) -> None:
        pos = self._toPoint(tup)
        self.firstS.labelMass.setPos(pos)

    @staticmethod
    def _toPoint(tup):
        pos = QPointF(tup[0], tup[1])
        return pos

    def setStartPort(self, newStartPort):
        self.fromPort = newStartPort
        self.startPos = newStartPort.scenePos()
        # self.fromPort.posCallbacks.append(self.setStartPos)

    def setEndPort(self, newEndPort):
        self.toPort = newEndPort

    def setStartPos(self):
        pass

    def setEndPos(self):
        pass

    def setColor(self, value, **kwargs):
        col = QColor(0, 0, 0)

        if "mfr" in kwargs:
            if kwargs["mfr"] == "NegMfr":
                col = QColor(0, 0, 255)
            elif kwargs["mfr"] == "ZeroMfr":
                col = QColor(142, 142, 142)  # Gray
            elif kwargs["mfr"] == "min":
                col = QColor(0, 0, 204)  # deep blue
            elif kwargs["mfr"] == "max":
                col = QColor(153, 0, 0)  # deep red
            elif kwargs["mfr"] == "minTo25":
                col = QColor(0, 128, 255)  # blue
            elif kwargs["mfr"] == "25To50":
                col = QColor(102, 255, 255)  # light blue
            elif kwargs["mfr"] == "50To75":
                col = QColor(255, 153, 153)  # light red
            elif kwargs["mfr"] == "75ToMax":
                col = QColor(255, 51, 51)  # red
            # elif kwargs["mfr"] == "negMinToLower":
            #     col = QColor(0, 0, 0)  # Black
            # elif kwargs["mfr"] == "negLowerToMedian":
            #     col = QColor(47, 47, 73)  # Lighter black
            # elif kwargs["mfr"] == "negMedianToUpper":
            #     col = QColor(78, 78, 97)  # even lighter black
            # elif kwargs["mfr"] == "negUpperToMax":
            #     col = QColor(100, 100, 114)  # even even lighter black
            # elif kwargs["mfr"] == "negMaxToZero":
            #     col = QColor(115, 115, 124)  # darker than gray
            # elif kwargs["mfr"] == "zeroToMin":
            #     col = QColor(156, 156, 164)  # lighter than gray
            else:
                # PosMfr
                col = QColor(255, 0, 0)

            for s in self.segments:
                self.logger.debug("Value: " + str(value))
                pen1 = QPen(col, value)
                s.setPen(pen1)

        else:
            self.logger.debug("No color to set in Connection.setColor()")

    # Getters
    def getStartPoint(self):
        return QPointF(self.fromPort.scenePos())

    def getEndPoint(self):
        return QPointF(self.toPort.scenePos())

    def getNodePos(self, searchCorner):

        corners = self.getCorners()

        for i in range(len(corners)):
            if corners[i] == searchCorner:
                # self.logger.debug("Found a corner")
                return i
        self.logger.debug("corner not found is " + str(searchCorner))

        return 0
        # Alternative
        # return corners.index(searchCorner)

    def getFirstSeg(self):
        return self.segments[0]

    def getCorners(self):
        # Returns a list containing all the corners of this connection
        res = []

        tempNode = self.startNode.nextN()

        while tempNode.nextN() is not None:
            res.append(tempNode.parent)
            tempNode = tempNode.nextN()

        # self.logger.debug("getcorners gives " + str(res))
        return res

    # To correct the direction of a connection, unused
    def correctPorts(self):
        if isinstance(self.fromPort.parent, Pump):
            if self.fromPort.name == "i":
                self.switchPorts()

        if isinstance(self.toPort.parent, Pump):
            if self.toPort.name == "o":
                self.switchPorts()

        if isinstance(self.fromPort, Collector):
            if self.fromPort.name == "i":
                self.switchPorts()

        if isinstance(self.toPort, Collector):
            if self.toPort.name == "o":
                self.switchPorts()

        if isinstance(self.fromPort.parent, TVentil):
            self.switchPorts()

    def switchPorts(self):
        temp = self.fromPort
        self.fromPort = self.toPort
        self.toPort = temp

    # Initialization
    def initNew(self, parent):

        self.parent = parent
        # self.logger.debug("Parent is " + str(self.parent))

        # Global
        self.parent.trnsysObj.append(self)

        self.displayName = "Conn" + str(self.connId)

        # global editorMode
        if self.parent.editorMode == 0:
            self.logger.debug("Creating a new connection in mode 0")
            self.initSegmentM0()
        elif self.parent.editorMode == 1:
            self.logger.debug("Creating a new connection in mode 1")
            self.initSegmentM1()
        else:
            self.logger.debug("No valid mode during creating of connection")

        self.parent.connectionList.append(self)
        self.fromPort.connectionList.append(self)
        self.toPort.connectionList.append(self)

        # self.correctPorts()

    def initLoad(self):
        # Called by DiagramEditor when loading connection

        self.logger.debug("Port 1 is " + str(self.fromPort) + "has Id" + str(self.fromPort.id))
        self.logger.debug("Port 2 is " + str(self.toPort))

        # self.parent = loadParent
        self.initSegmentM1()

        self.parent.trnsysObj.append(self)

        self.parent.connectionList.append(self)
        self.fromPort.connectionList.append(self)
        self.toPort.connectionList.append(self)

        if len(self.cornersLoad) > 0:
            self.loadSegments()

        if self.labelPosLoad:
            self.setLabelPos(self.labelPosLoad)

        if self.labelMassPosLoad:
            self.setMassLabelPos(self.labelMassPosLoad)

        # Still not tested
        # self.correctPorts()

    def initSegmentM0(self):
        self.startNode.setParent(self)
        self.endNode.setParent(self)

        self.startNode.setNext(self.endNode)
        self.endNode.setPrev(self.startNode)

        self.firstS = segmentItem(self.startNode, self.endNode, self)

        self.firstS.setLine(QLineF(self.getStartPoint(), self.getEndPoint()))

        self.parent.diagramScene.addItem(self.firstS)

        self.positionLabel()

    def initSegmentM1(self):
        # Can be rewritten for efficiency etc. (e.g not doing initSegmentM0 and calling niceConn())

        self.startNode.setParent(self)
        self.endNode.setParent(self)

        self.startNode.setNext(self.endNode)
        self.endNode.setPrev(self.startNode)

        self.firstS = segmentItem(self.startNode, self.endNode, self)

        self.firstS.setLine(QLineF(self.getStartPoint(), self.getEndPoint()))

        self.parent.diagramScene.addItem(self.firstS)

        self.niceConn()

        self.positionLabel()

    def loadSegments(self):
        self.clearConn()

        self.logger.debug("cornerLoads is " + str(self.cornersLoad))

        tempNode = self.startNode

        self.logger.debug("start node has nn " + str(tempNode.nextNode))
        self.logger.debug(" ")
        rad = 2

        for x in range(len(self.cornersLoad)):
            cor = CornerItem(-rad, -rad, 2 * rad, 2 * rad, tempNode, tempNode.nextN(), self)
            # self.logger.debug("items are now " + str(self.parent.diagramScene.items()))
            self.logger.debug("Corner" + str(cor) + " has  " + str(cor.node))

            cor.setPos(float(self.cornersLoad[x][0]), float(self.cornersLoad[x][1]))
            # cor.setFlags(cor.ItemIsSelectable | cor.ItemIsMovable)
            cor.setFlag(cor.ItemSendsScenePositionChanges, True)

            cor.setZValue(100)
            cor.setVisible(True)

            self.parent.diagramScene.addItem(cor)

            cor.node.nextNode.setPrev(cor.node)
            cor.node.prevNode.setNext(cor.node)
            segmentItem(tempNode, cor.node, self)

            tempNode = cor.node

            self.printConn()

        segmentItem(tempNode, tempNode.nextN(), self)

        # self.parent.diagramScene.addItem(lastSeg)

        self.printConn()

        for s in self.segments:
            pos1 = None
            pos2 = None

            if isinstance(s.startNode.parent, Connection) and s.startNode.prevNode is None:
                # self.logger.debug("startnode is at connection")
                pos1 = s.startNode.parent.fromPort.scenePos().x(), s.startNode.parent.fromPort.scenePos().y()
            if isinstance(s.endNode.parent, Connection) and s.endNode.nextNode is None:
                # self.logger.debug("endnode is at connection")
                pos2 = s.endNode.parent.toPort.scenePos().x(), s.endNode.parent.toPort.scenePos().y()

            if type(s.startNode.parent) is CornerItem:
                # self.logger.debug("startnode is at corner")
                self.logger.debug(
                    str(s.startNode.parent) + " " + str(s.startNode) + "cor " + str(s.startNode.parent.scenePos())
                )
                pos1 = s.startNode.parent.scenePos().x(), s.startNode.parent.scenePos().y()
            if type(s.endNode.parent) is CornerItem:
                # self.logger.debug("endnode is at corner")
                pos2 = s.endNode.parent.scenePos().x(), s.endNode.parent.scenePos().y()

            self.logger.debug("pos1 is " + str(pos1))
            self.logger.debug("pos2 is " + str(pos2))

            s.setLine(QLineF(pos1[0], pos1[1], pos2[0], pos2[1]))

            self.parent.diagramScene.addItem(s)

        self.firstS = self.segments[0]

        self.positionLabel()

    # Label related
    def setLabelVisible(self, isVisible: bool) -> None:
        self.firstS.setLabelVisible(isVisible)

    def toggleLabelVisible(self) -> None:
        self.firstS.toggleLabelVisible()

    def updateSegLabels(self):
        for s in self.segments:
            s.label.setPlainText(self.displayName)

    def positionLabel(self):

        if self.parent.editorMode == 0:
            vec1 = self.firstS.line().p2() - self.firstS.line().p1()
            vec2 = QPointF(1 / 10 * vec1.x(), 1 / 10 * vec1.y())

            eps = 0.5
            if vec1.x() == 0:
                angleBetween = _math.atan(vec1.y() / eps)
            else:
                angleBetween = _math.atan(vec1.y() / vec1.x())

            d1 = calcDist(QPointF(0, 0), vec1)

            if d1 == 0:
                d1 = eps

            vec3 = QPointF(-5 * angleBetween * vec1.y() / d1, 5 * angleBetween * vec1.x() / d1)
            self.firstS.label.setPos(self.fromPort.pos() + vec2 + vec3)

        elif self.parent.editorMode == 1:
            # self.logger.debug("changing label pos in editor mode 1")
            if self.fromPort.side == 0:
                self.firstS.label.setPos(QPointF(-60, 0))

            # Here the behavior of positioning can be improved
            elif self.fromPort.side in [1, 2, 3]:
                self.firstS.label.setPos(QPointF(20, 0))
        else:
            pass
        if self.segmentsLoad:
            vec1 = np.array(self.segmentsLoad[0][:2])
            vec2 = np.array(self.segmentsLoad[0][2:])
            vec = vec2 - vec1
            # uVec1 = vec1/np.linalg.norm(vec1)
            # uVec2 = vec2/np.linalg.norm(vec2)
            uVec = vec / np.linalg.norm(vec)
            dotProduct = np.dot(np.array([1.0, 0]), uVec)
            angle = np.rad2deg(np.arccos(dotProduct))
            if np.isnan(angle):
                angle = 0
            self.firstS.label.setRotation(-angle)

    def setMassFlowLabelVisible(self, isVisible: bool) -> None:
        self.firstS.setMassFlowLabelVisible(isVisible)

    def toggleMassFlowLabelVisible(self) -> None:
        self.firstS.toggleMassFlowLabelVisible()

    # Makes 90deg angles of connection
    def niceConn(self):
        """
        Creates the segments and corners depending on the side of the fromPort and toPort
        Returns
        -------

        """
        # Here different cases can be implemented using self.PORT.side as sketched on paper
        rad = 2  # 4

        self.logger.debug(
            "FPort " + str(self.fromPort) + " has side " + str(self.fromPort.side) + " has " + str(self.fromPort.name)
        )
        self.logger.debug(
            "FPort " + str(self.fromPort) + " has side " + str(self.fromPort.side) + " has " + str(self.fromPort.name)
        )

        if (self.fromPort.side == 2) and (self.toPort.side == 2):
            self.fromPort.createdAtSide = 2
            self.toPort.createdAtSide = 2
            self.logger.debug("NiceConn 2 to 2")
            portOffset = 30
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, None, self)
            corner3 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner2.node, None, self)
            corner4 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner3.node, self.endNode, self)

            corner1.node.setNext(corner2.node)
            corner2.node.setNext(corner3.node)
            corner3.node.setNext(corner4.node)

            seg1 = segmentItem(self.startNode, corner1.node, self)
            seg2 = segmentItem(corner1.node, corner2.node, self)
            seg3 = segmentItem(corner2.node, corner3.node, self)
            seg4 = segmentItem(corner3.node, corner4.node, self)
            seg5 = segmentItem(corner4.node, self.endNode, self)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner4.node)

            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)
            self.parent.diagramScene.addItem(seg4)
            self.parent.diagramScene.addItem(seg5)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)
            self.parent.diagramScene.addItem(corner3)
            self.parent.diagramScene.addItem(corner4)

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            baseline_h = max(pos1.y(), pos2.y()) + 100.6

            p1 = QPointF(pos1.x() + portOffset, pos1.y())
            p2 = QPointF(p1.x(), baseline_h)
            p3 = QPointF(pos2.x() + portOffset, baseline_h)
            p4 = QPointF(p3.x(), pos2.y())

            seg1.setLine(QLineF(pos1, p1))
            seg2.setLine(QLineF(p1, p2))
            seg3.setLine(QLineF(p2, p3))
            seg4.setLine(QLineF(p3, p4))
            seg5.setLine(QLineF(p4, pos2))

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)
            corner3.setFlag(corner3.ItemSendsScenePositionChanges, True)
            corner4.setFlag(corner4.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            corner3.setZValue(100)
            corner4.setZValue(100)

            corner1.setPos(p1)
            corner2.setPos(p2)
            corner3.setPos(p3)
            corner4.setPos(p4)

        elif (self.fromPort.side == 0) and (self.toPort.side == 0):
            self.fromPort.createdAtSide = 0
            self.toPort.createdAtSide = 0
            self.logger.debug("NiceConn 0 to 0")
            portOffset = 30
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, None, self)
            corner3 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner2.node, None, self)
            corner4 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner3.node, self.endNode, self)

            corner1.node.setNext(corner2.node)
            corner2.node.setNext(corner3.node)
            corner3.node.setNext(corner4.node)

            seg1 = segmentItem(self.startNode, corner1.node, self)
            seg2 = segmentItem(corner1.node, corner2.node, self)
            seg3 = segmentItem(corner2.node, corner3.node, self)
            seg4 = segmentItem(corner3.node, corner4.node, self)
            seg5 = segmentItem(corner4.node, self.endNode, self)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner4.node)

            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)
            self.parent.diagramScene.addItem(seg4)
            self.parent.diagramScene.addItem(seg5)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)
            self.parent.diagramScene.addItem(corner3)
            self.parent.diagramScene.addItem(corner4)

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            baseline_h = max(pos1.y(), pos2.y()) + 100.6

            # p1 = QPointF(pos1.x() + 50, pos1.y())
            p1 = QPointF(pos1.x() - portOffset, pos1.y())
            p2 = QPointF(p1.x(), baseline_h)
            p3 = QPointF(pos2.x() - portOffset, baseline_h)
            p4 = QPointF(p3.x(), pos2.y())

            seg1.setLine(QLineF(pos1, p1))
            seg2.setLine(QLineF(p1, p2))
            seg3.setLine(QLineF(p2, p3))
            seg4.setLine(QLineF(p3, p4))
            seg5.setLine(QLineF(p4, pos2))

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)
            corner3.setFlag(corner3.ItemSendsScenePositionChanges, True)
            corner4.setFlag(corner4.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            corner3.setZValue(100)
            corner4.setZValue(100)

            corner1.setPos(p1)
            corner2.setPos(p2)
            corner3.setPos(p3)
            corner4.setPos(p4)

        elif self.fromPort.side == 1:
            self.fromPort.createdAtSide = 1
            # todo :  when rotated, it cause a problem because side gets changed

            self.logger.debug("NiceConn from 1")
            portOffset = 30
            self.clearConn()

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            if pos2.y() <= pos1.y():
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, self.endNode, self)

                seg1 = segmentItem(self.startNode, corner1.node, self)
                seg2 = segmentItem(corner1.node, self.endNode, self)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner1.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(corner1)

                p1 = QPointF(pos1.x(), pos2.y() - 0.333)  # position of the connecting node

                seg1.setLine(QLineF(pos1, p1))
                seg2.setLine(QLineF(p1, pos2))

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner1.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)

                corner1.setPos(p1)
                self.firstS = self.getFirstSeg()

            else:
                self.logger.debug("To port below from port")
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
                corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

                corner1.node.setNext(corner2.node)

                seg1 = segmentItem(self.startNode, corner1.node, self)
                seg2 = segmentItem(corner1.node, corner2.node, self)
                seg3 = segmentItem(corner2.node, self.endNode, self)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner2.node)

                self.printConnNodes()
                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(seg3)

                self.parent.diagramScene.addItem(corner1)
                self.parent.diagramScene.addItem(corner2)

                offsetPoint = pos1.y() - 15.666

                help_point_1 = QPointF(pos1.x(), offsetPoint)
                help_point_2 = QPointF(pos2.x(), offsetPoint)

                seg1.setLine(QLineF(pos1, help_point_1))
                seg2.setLine(QLineF(help_point_1, help_point_2))
                seg3.setLine(QLineF(help_point_2, pos2))

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

                corner1.setZValue(100)
                corner2.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)
                self.logger.debug("Here in niceconn")

                corner1.setPos(help_point_1)
                corner2.setPos(help_point_2)
                self.firstS = self.getFirstSeg()

        elif self.fromPort.side == 3:
            self.fromPort.createdAtSide = 3

            self.logger.debug("NiceConn from 1")
            portOffset = 30
            self.clearConn()

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            if pos2.y() >= pos1.y():
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, self.endNode, self)

                seg1 = segmentItem(self.startNode, corner1.node, self)
                seg2 = segmentItem(corner1.node, self.endNode, self)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner1.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(corner1)

                p1 = QPointF(pos1.x(), pos2.y() - 0.333)  # position of the connecting node

                seg1.setLine(QLineF(pos1, p1))
                seg2.setLine(QLineF(p1, pos2))

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner1.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)

                corner1.setPos(p1)
                self.firstS = self.getFirstSeg()
            else:
                self.logger.debug("To port above from port")
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
                corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

                corner1.node.setNext(corner2.node)

                seg1 = segmentItem(self.startNode, corner1.node, self)
                seg2 = segmentItem(corner1.node, corner2.node, self)
                seg3 = segmentItem(corner2.node, self.endNode, self)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner2.node)

                self.printConnNodes()
                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(seg3)

                self.parent.diagramScene.addItem(corner1)
                self.parent.diagramScene.addItem(corner2)

                offsetPoint = pos1.y() + 15.666

                help_point_1 = QPointF(pos1.x(), offsetPoint)
                help_point_2 = QPointF(pos2.x(), offsetPoint)

                seg1.setLine(QLineF(pos1, help_point_1))
                seg2.setLine(QLineF(help_point_1, help_point_2))
                seg3.setLine(QLineF(help_point_2, pos2))

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

                corner1.setZValue(100)
                corner2.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)
                self.logger.debug("Here in niceconn")

                corner1.setPos(help_point_1)
                corner2.setPos(help_point_2)
                self.firstS = self.getFirstSeg()

        else:
            # if((self.fromPort.side == 2) and (self.toPort.side == 0)) or (
            #         (self.fromPort.side == 0) and (self.toPort.side == 2) or (self.fromPort.side == 1) and (
            #         self.toPort.side in [0, 1, 2]) or (self.fromPort.side in [0, 1, 2]) and (self.toPort.side == 1)):

            self.fromPort.createdAtSide = self.fromPort.side
            self.toPort.createdAtSide = self.toPort.side
            self.logger.debug("Ports are directed to each other")
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

            corner1.node.setNext(corner2.node)

            seg1 = segmentItem(self.startNode, corner1.node, self)
            seg2 = segmentItem(corner1.node, corner2.node, self)
            seg3 = segmentItem(corner2.node, self.endNode, self)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner2.node)

            # self.logger.debug("niceConn...")
            self.printConnNodes()
            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            midx = pos1.x() + 0.5 * (pos2.x() - pos1.x())

            help_point_1 = QPointF(midx, pos1.y())
            help_point_2 = QPointF(midx, pos2.y())

            seg1.setLine(QLineF(pos1, help_point_1))
            seg2.setLine(QLineF(help_point_1, help_point_2))
            seg3.setLine(QLineF(help_point_2, pos2))

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            self.fromPort.setZValue(100)
            self.toPort.setZValue(100)
            self.logger.debug("Here in niceconn")

            corner1.setPos(help_point_1)
            corner2.setPos(help_point_2)
            self.firstS = self.getFirstSeg()

            self.logger.debug("Conn has now " + str(self.firstS))

    # Unused
    def buildBridges(self):
        # This function finds the colliding line segments and creates the interrupted line effect

        # Debug:
        # self.logger.debug("Segments in brigdes function is " + str(self.segments))
        # self.logger.debug("Length of segments is " + str(len(self.segments)))
        for s in self.segments:
            col = s.collidingItems()
            # Why do the child segments have again colliding Items?
            # self.logger.debug("Start and end of loop segment is " + str(s.startNode) + " and " + str(s.endNode))
            # self.logger.debug("Col items are " + str(col))

            for c in col:
                if type(c) is segmentItem:
                    # self.logger.debug("There is a segment colliding....")
                    # self.logger.debug("c.hasBridge is " + str(c.hasBridge))
                    # self.logger.debug("s.hasBridge is " + str(s.hasBridge))

                    # Both have no bridge and do not collide at the endpoints:
                    # if (s.hasBridge == False) and (c.hasBridge == False) and (c.endNode is not s.startNode) and (c.startNode is not s.endNode):
                    if (c.endNode is not s.startNode) and (c.startNode is not s.endNode):

                        # self.logger.debug("Self has startnode " + str(s.startNode) + " and endnode " + str(
                        #     s.endNode) + " c has startNode " + str(c.startNode) + " and endnode " + str(c.endNode))

                        qp1 = s.line().p1()
                        qp2 = s.line().p2()
                        qp1_ = c.line().p1()
                        qp2_ = c.line().p2()

                        eps2 = 5
                        if abs(qp1.x() - qp2.x()) < eps2 or abs(qp1_.x() - qp2_.x()) < eps2:
                            self.logger.debug("Can't build bridge because one segment is almost verical")

                        else:

                            node1 = Node()
                            node2 = Node()

                            node1.setPrev(s.startNode)
                            node1.setNext(node2)
                            node2.setPrev(node1)
                            node2.setNext(s.endNode)
                            node1.setParent(self)
                            node2.setParent(self)

                            s.startNode.setNext(node1)
                            s.endNode.setPrev(node2)

                            s.firstChild = segmentItem(s.startNode, node1, s.parent)
                            s.secondChild = segmentItem(node2, s.endNode, s.parent)

                            s.hasBridge = True
                            c.hasBridge = True
                            s.bridgedSegment = c
                            c.bridgedSegment = s

                            s.firstChild.setVisible(False)
                            s.secondChild.setVisible(False)

                            self.parent.diagramScene.addItem(s.firstChild)
                            self.parent.diagramScene.addItem(s.secondChild)

                            # self.logger.debug("Points are " + str(qp1) + str(qp2) + str(qp1_) + str(qp2_))
                            # eps1 = 1
                            # if (abs(qp2.x() - qp1.x()) > eps1) and (abs(qp2_.x() - qp1_.x()) > eps1) :

                            n1 = (qp1.y() * qp2.x() - qp2.y() * qp1.x()) / (qp2.x() - qp1.x())
                            n2 = (qp1_.y() * qp2_.x() - qp2_.y() * qp1_.x()) / (qp2_.x() - qp1_.x())

                            m1 = (qp2.y() - qp1.y()) / (qp2.x() - qp1.x())
                            m2 = (qp2_.y() - qp1_.y()) / (qp2_.x() - qp1_.x())

                            # self.logger.debug("division by " + str(m1) + "    " + str(m2))
                            collisionPos = QPointF((n1 - n2) / (m2 - m1), m1 * (n1 - n2) / (m2 - m1) + n1)

                            vecp2p1 = qp2 - qp1
                            vecp2p1_ = qp2_ - qp1_

                            normVec = _math.sqrt(vecp2p1.x() ** 2 + vecp2p1.y() ** 2)
                            normVec_ = _math.sqrt(vecp2p1_.x() ** 2 + vecp2p1_.y() ** 2)

                            # Compute angle between lines
                            scalarProd = vecp2p1.x() * vecp2p1_.x() + vecp2p1.y() + vecp2p1_.y()
                            angleBetween = _math.acos(scalarProd / (normVec * normVec_))
                            # self.logger.debug(str(degrees(angleBetween)))

                            # Almost working: if line approaches 90 deg or if lines touch because too steep, problem
                            f = 10
                            distFactor = max(f * 1 / angleBetween, f)
                            # self.logger.debug("distFactor " +  str(distFactor))

                            # normalVec = QPointF(-distFactor/normVec * vecp2p1.y(), distFactor/normVec * vecp2p1.x())
                            normVecp2p1 = QPointF(distFactor / normVec * vecp2p1.x(), distFactor / normVec * vecp2p1.y())

                            # self.logger.debug("vector is " + str(normVecp2p1))
                            # self.logger.debug("collisionpos is " + str(collisionPos))
                            # newPos1 = collisionPos - normalVec
                            # newPos2 = collisionPos + normalVec

                            newPos1 = collisionPos - normVecp2p1
                            newPos2 = collisionPos + normVecp2p1

                            s.firstChild.setLine(QLineF(qp1.x(), qp1.y(), newPos1.x(), newPos1.y()))
                            s.secondChild.setLine(QLineF(newPos2.x(), newPos2.y(), qp2.x(), qp2.y()))

                            s.firstChild.setVisible(True)
                            s.secondChild.setVisible(True)

                            s.hide()

                            self.parent.diagramScene.removeItem(s)
                            self.segments.remove(s)

                            # self.logger.debug("s.bridged element is " + str(s.bridgedSegment))
                            # self.logger.debug("c.bridged element is " + str(c.bridgedSegment))
                            # self.logger.debug(len(self.segments))

    # Delete a connection
    def clearConn(self):
        # Deletes all segments and corners in connection
        # self.logger.debug("Connection was before: ")
        # self.printConnNodes()

        walker = self.endNode

        items = self.parent.diagramScene.items()

        for it in items:
            if type(it) is segmentItem:
                if it.startNode.lastNode() is self.endNode or it.startNode.firstNode() is self.startNode:
                    # self.logger.debug("Del segment")
                    self.parent.diagramScene.removeItem(it)
            if type(it) is QGraphicsTextItem:
                if type(it.parent) is PortItem:
                    self.logger.debug("it has " + str(it.parent()))
                    self.logger.debug("Deleting it")
                    self.parent.diagramScene.removeItem(it)

        for seg in self.segments:
            self.parent.diagramScene.removeItem(seg.label)

        self.segments.clear()

        if walker.prevN() is not self.startNode:
            walker = walker.prevN()

            while walker.prevN() is not self.startNode:
                # self.logger.debug("Del corner...")
                if type(walker.parent) is not Connection:
                    self.parent.diagramScene.removeItem(walker.parent)
                else:
                    self.logger.debug("Caution, this is a disrupt.")
                walker = walker.prevN()
                # del (walker.nextN())

            if type(walker.parent) is not Connection:
                self.parent.diagramScene.removeItem(walker.parent)
            else:
                self.logger.debug("Caution.")
                # del walker

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

        # self.logger.debug("Connection is now: ")
        # self.printConnNodes()

    def deleteConn(self):
        self.logger.debug("Deleting connection " + self.displayName + " " + str(self))
        self.logger.debug("fromPort is at " + str(self.fromPort.parent.displayName))
        self.logger.debug("toPort is at " + str(self.toPort.parent.displayName))

        self.clearConn()
        self.fromPort.connectionList.remove(self)

        # self.logger.debug("Connectionlist of fromPort after removal is:")
        # [self.logger.debug(c.displayName + ": from " + str(c.fromPort.parent) + " to " + str(c.toPort.parent)) for c in self.fromPort.connectionList ]

        self.toPort.connectionList.remove(self)

        # self.logger.debug("Connectionlist of toPort after removal is:")
        # [self.logger.debug(c.displayName + ": from " + str(c.fromPort.parent) + " to " + str(c.toPort.parent)) for c in self.toPort.connectionList ]

        if self in self.parent.trnsysObj:
            self.parent.trnsysObj.remove(self)
        else:
            self.logger.debug("-------> tr obj are " + str(self.parent.trnsysObj))
            self.logger.debug(self.id)
            self.logger.debug([i.id for i in self.parent.trnsysObj])
            return

        self.removeConnFromGroup()

        self.logger.debug("Removing trnsysObj " + str(self))
        self.parent.connectionList.remove(self)
        del self

    def deleteConnCom(self):
        command = DeleteConnectionCommand(self, "Delete conn comand")
        self.parent.parent().undoStack.push(command)

    # Gradient related
    def totalLength(self):
        s = self.startNode
        e = self.endNode
        res = 0

        for seg in self.segments:
            res += calcDist(seg.line().p1(), seg.line().p2())
        return res

    def partialLength(self, node):
        # Returns the cummulative length of line up to given node
        # Assumes that segments is ordered correctly!
        res = 0
        # self.logger.debug("Segments has len " + str(len(self.segments)))
        if node == self.startNode:
            return res

        for i in self.segments:
            # self.logger.debug("Adding " + str(calcDist(i.line().p1(), i.line().p2())))
            res += calcDist(i.line().p1(), i.line().p2())
            if i.endNode == node:
                # self.logger.debug("Breaking")
                break

        # self.logger.debug("Node " + str(node) + " has pL " + str(res) + "\n")
        return res

    def updateSegGrads(self):
        for s in self.segments:
            s.updateGrad()

    # Invert connection
    def invertConnection(self):
        # Invert segment list
        self.segments.reverse()

        # Invert nodes
        self.invertNodes()

        # Invert ports
        temp = self.toPort
        self.toPort = self.fromPort
        self.fromPort = temp

        sn = self.startNode
        self.startNode = self.endNode
        self.endNode = sn

        for s in self.segments:
            temp2 = s.startNode
            s.startNode = s.endNode
            s.endNode = temp2

            temp3 = s.firstChild
            s.firstChild = s.secondChild
            s.secondChild = temp3

            s.setLine(s.line().p2().x(), s.line().p2().y(), s.line().p1().x(), s.line().p1().y())

            s.updateGrad()

    def invertNodes(self):
        temp = None
        element = self.startNode

        while element.nextN() is not None:
            temp = element.nextN()
            pr = element.prevN()
            ne = element.nextN()

            element.setNext(pr)
            element.setPrev(ne)
            element = temp

        pr = element.prevN()
        ne = element.nextN()
        element.setNext(pr)
        element.setPrev(ne)

    # Group related
    def setDefaultGroup(self):
        self.setConnToGroup("defaultGroup")

    def setConnToGroup(self, newGroupName):
        self.logger.debug("In setConnToGroup")
        if newGroupName == self.groupName:
            self.logger.debug("Block " + str(self) + str(self.displayName) + "is already in this group")
            return
        else:
            # self.logger.debug("groups is " + str(self.parent.groupList))
            for g in self.parent.groupList:
                self.logger.debug("At group " + str(g.displayName))
                if g.displayName == self.groupName:
                    self.logger.debug("Found the old group")
                    g.itemList.remove(self)
                if g.displayName == newGroupName:
                    self.logger.debug("Found the new group")
                    g.itemList.append(self)

            self.groupName = newGroupName

    def removeConnFromGroup(self):
        for g in self.parent.groupList:
            if g.displayName == self.groupName:
                if self in g.itemList:
                    g.itemList.remove(self)
                else:
                    self.logger.debug("While removing conn from group, groupName is invalid")
            else:
                self.logger.debug("While removeing conn from group, no group with conn.groupName")

    # Highlight when clicked, unhighlight when clicked elsewhere
    def highlightConn(self):
        self.unhighlightOtherConns()

        for s in self.segments:
            s.setHighlight(True)

        self.setLabelsHighlight(True)

    def unhighlightOtherConns(self):
        for c in self.parent.connectionList:
            c.unhighlightConn()

    def unhighlightConn(self):
        for s in self.segments:
            s.updateGrad()

        self.setLabelsHighlight(False)

    def setLabelsHighlight(self, isHighlight: bool) -> None:
        self._setBold(self.firstS.label, isHighlight)
        self._setBold(self.firstS.labelMass, isHighlight)

    @staticmethod
    def _setBold(label: QGraphicsTextItem, isBold: bool) -> None:
        originalFontCopy = label.font()
        originalFontCopy.setBold(isBold)
        label.setFont(originalFontCopy)

    # Debug
    def inspectConn(self):
        self.parent.listV.clear()
        self.parent.listV.addItem("Display name: " + self.displayName)
        self.parent.listV.addItem("Group name: " + self.groupName)
        self.parent.listV.addItem("Parent: " + str(self.parent))
        # self.parent.listV.addItem("Position: " + str(self.pos()))
        # self.parent.listV.addItem("Sceneposition: " + str(self.scenePos()))
        self.parent.listV.addItem("fromPort: " + str(self.fromPort) + str(self.fromPort.id))
        self.parent.listV.addItem("toPort: " + str(self.toPort) + str(self.toPort.id))

    def printConn(self):
        self.logger.debug("------------------------------------")
        self.logger.debug("This is the printout of connection: " + self.displayName + str(self))
        self.logger.debug("It has fromPort " + str(self.fromPort))
        self.logger.debug("It has toPort " + str(self.toPort))
        self.logger.debug("It has startNode " + str(self.startNode))
        self.logger.debug("It has endNode " + str(self.endNode))

        # self.printConnNodes()
        self.logger.debug("\n")
        # self.printConnSegs()
        self.logger.debug("It goes from " + self.fromPort.parent.displayName + " to " + self.toPort.parent.displayName)
        self.logger.debug("------------------------------------")

    def printConnNodes(self):
        self.logger.debug("These are the nodes: ")

        element = self.startNode
        while element.nextN() is not None:
            self.logger.debug(
                "Node is "
                + str(element)
                + " has nextNode "
                + str(element.nextN())
                + " has pL "
                + str(self.partialLength(element))
            )
            element = element.nextN()

        self.logger.debug("Node is " + str(element) + " has nextNode " + str(element.nextN()))

    def printConnSegs(self):
        self.logger.debug("These are the segments in order")
        for s in self.segments:
            self.logger.debug(
                "Segment is "
                + str(s)
                + " has global id "
                + str(s.id)
                + " has startnode "
                + str(s.startNode)
                + " endnode "
                + str(s.endNode)
            )

    # Saving / Loading
    def encode(self):
        self.logger.debug("Encoding a connection")

        dct = {}
        dct[".__ConnectionDict__"] = True
        dct["PortFromID"] = self.fromPort.id
        dct["PortToID"] = self.toPort.id
        dct["ConnDisplayName"] = self.displayName
        dct["ConnID"] = self.id
        dct["ConnCID"] = self.connId
        dct["trnsysID"] = self.trnsysId
        dct["GroupName"] = self.groupName

        segments = []  # Not used, but instead corners[]

        for s in self.segments:
            segmentTupel = (s.line().p1().x(), s.line().p1().y(), s.line().p2().x(), s.line().p2().y())
            segments.append(segmentTupel)
        # self.logger.debug("Segments in encoder is " + str(segments))
        dct["SegmentPositions"] = segments
        if len(self.segments) > 0:
            dct["FirstSegmentLabelPos"] = self.segments[0].label.pos().x(), self.segments[0].label.pos().y()
            dct["FirstSegmentMassFlowLabelPos"] = (
                self.segments[0].labelMass.pos().x(),
                self.segments[0].labelMass.pos().y(),
            )
        else:
            self.logger.debug("This connection has no segment")
            defaultPosition = self.fromPort.pos().x(), self.fromPort.pos().y()
            dct["FirstSegmentLabelPos"] = defaultPosition
            dct["FirstSegmentMassFlowLabelPos"] = defaultPosition

        corners = []

        for s in self.getCorners():
            cornerTupel = (s.pos().x(), s.pos().y())
            corners.append(cornerTupel)
        dct["CornerPositions"] = corners
        dictName = "Connection-"

        return dictName, dct

    def decode(self, i):
        self.logger.debug("Loading a connection in Decoder")

        self.id = i["ConnID"]
        self.connId = i["ConnCID"]
        self.trnsysId = i["trnsysID"]
        self.setName(i["ConnDisplayName"])
        self.groupName = "defaultGroup"
        self.setConnToGroup(i["GroupName"])

    # Export related
    def exportBlackBox(self):
        return "noBlackBoxOutput", []

    def exportPumpOutlets(self):
        return "", 0

    def exportMassFlows(self):
        return "", 0

    def exportDivSetting1(self):
        return "", 0

    def exportDivSetting2(self, nUnit):
        return "", nUnit

    def exportParametersFlowSolver(self, descConnLength):
        f = ""

        if hasattr(self.fromPort.parent, "getSubBlockOffset"):
            temp = str(self.fromPort.parent.trnsysId + self.fromPort.parent.getSubBlockOffset(self)) + " "
        else:
            portId = self.fromPort.parent.getFlowSolverParametersId(self.fromPort)
            temp = f"{portId} "

        if hasattr(self.toPort.parent, "getSubBlockOffset"):
            temp += str(self.toPort.parent.trnsysId + self.toPort.parent.getSubBlockOffset(self))
        else:
            portId = self.toPort.parent.getFlowSolverParametersId(self.toPort)
            temp += f"{portId} "

        temp += " 0" * 2 + " "
        self.exportConnsString = temp

        # This is to ensure that the "output" of a Div always appears first
        fromPortWithParent = (self.fromPort, self.fromPort.parent)
        if type(self.fromPort.parent) is TVentil and self.fromPort in self.fromPort.parent.outputs:
            self._portItemsWithParent.insert(0, fromPortWithParent)
        else:
            self._portItemsWithParent.append(fromPortWithParent)

        toPortWithParent = (self.toPort, self.toPort.parent)
        if type(self.toPort.parent) is TVentil and self.fromPort in self.toPort.parent.outputs:
            self._portItemsWithParent.insert(0, toPortWithParent)
        else:
            self._portItemsWithParent.append(toPortWithParent)

        f += temp + " " * (descConnLength - len(temp)) + "!" + str(self.trnsysId) + " : " + str(self.displayName) + "\n"

        return f

    @staticmethod
    def exportInputsFlowSolver1():
        return "0,0 ", 1

    def exportInputsFlowSolver2(self):
        return str(self.exportInitialInput) + " ", 1

    def exportOutputsFlowSolver(self, prefix, abc, equationNumber, simulationUnit):
        equation1 = self._createFlowSolverOutputEquation(0, abc, prefix, equationNumber, simulationUnit)
        equation2 = self._createFlowSolverOutputEquation(1, abc, prefix, equationNumber, simulationUnit)

        self.exportEquations.append(equation1)
        self.exportEquations.append(equation2)

        equations = equation1 + equation2
        nEquationsUsed = 2
        nextEquationNumber = equationNumber + 3

        return equations, nextEquationNumber, nEquationsUsed

    def _createFlowSolverOutputEquation(self, equationNumber, abc, prefix, equationNumberOffset, simulationUnit):
        return f"{prefix}{self.displayName}_{abc[equationNumber]}=[{simulationUnit},{equationNumberOffset + equationNumber}]\n"

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        f = ""
        unitNumber = startingUnit
        typeNr2 = 931  # Temperature calculation from a pipe

        unitText = ""
        ambientT = 20

        densityVar = "RhoWat"
        specHeatVar = "CPWat"

        equationConstant1 = 1
        equationConstant2 = 3

        parameterNumber = 6
        inputNumbers = 4

        # Fixed strings
        diameterPrefix = "di"
        lengthPrefix = "L"
        lossPrefix = "U"
        tempRoomVar = "TRoomStore"
        initialValueS = "20 0.0 20 20"
        powerPrefix = "P"

        # Momentarily hardcoded
        equationNr = 3

        unitText += "UNIT " + str(unitNumber) + " TYPE " + str(typeNr2) + "\n"
        unitText += "!" + self.displayName + "\n"
        unitText += "PARAMETERS " + str(parameterNumber) + "\n"

        unitText += diameterPrefix + self.displayName + "\n"
        unitText += lengthPrefix + self.displayName + "\n"
        unitText += lossPrefix + self.displayName + "\n"
        unitText += densityVar + "\n"
        unitText += specHeatVar + "\n"
        unitText += str(ambientT) + "\n"

        unitText += "INPUTS " + str(inputNumbers) + "\n"

        if len(self._portItemsWithParent) == 2:
            portItem = self._portItemsWithParent[0][0]
            parent = self._portItemsWithParent[0][1]
            if hasattr(parent, "getSubBlockOffset"):
                unitText += (
                    "T"
                    + parent.displayName
                    + "X"
                    + str(parent.getSubBlockOffset(self) + 1)
                    + "\n"
                )
            else:
                unitText += parent.getTemperatureVariableName(portItem) + "\n"

            unitText += self.exportEquations[0][0 : self.exportEquations[0].find("=")] + "\n"
            unitText += tempRoomVar + "\n"

            portItem = self._portItemsWithParent[1][0]
            parent = self._portItemsWithParent[1][1]
            if hasattr(parent, "getSubBlockOffset"):
                unitText += (
                    "T"
                    + parent.displayName
                    + "X"
                    + str(parent.getSubBlockOffset(self) + 1)
                    + "\n"
                )
            else:
                unitText += parent.getTemperatureVariableName(portItem) + "\n"

        else:
            f += (
                "Error: NO VALUE\n" * 3
                + "at connection with parents "
                + str(self.fromPort.parent)
                + str(self.toPort.parent)
                + "\n"
            )

        unitText += "***Initial values\n"
        unitText += initialValueS + "\n\n"

        unitText += "EQUATIONS " + str(equationNr) + "\n"
        unitText += "T" + self.displayName + "= [" + str(unitNumber) + "," + str(equationConstant1) + "]\n"
        unitText += (
            powerPrefix
            + self.displayName
            + "_kW"
            + "= ["
            + str(unitNumber)
            + ","
            + str(equationConstant2)
            + "]/3600 !kW\n"
        )
        unitText += "Mfr" + self.displayName + "= " + "Mfr" + self.displayName + "_A" "\n"

        unitNumber += 1
        unitText += "\n"
        f += unitText

        return unitText, unitNumber

    def findStoragePort(self, virtualBlock):
        portToPrint = None
        for p in virtualBlock.inputs + virtualBlock.outputs:
            if self in p.connectionList:
                # Found the port of the generated block adjacent to this pipe
                # Assumes 1st connection is with storageTank
                if self.fromPort == p:
                    if self.toPort.connectionList[0].fromPort == self.toPort:
                        portToPrint = self.toPort.connectionList[0].toPort
                    else:
                        portToPrint = self.toPort.connectionList[0].fromPort
                else:
                    if self.fromPort.connectionList[0].fromPort == self.fromPort:
                        portToPrint = self.fromPort.connectionList[0].toPort
                    else:
                        portToPrint = self.fromPort.connectionList[0].fromPort
        return portToPrint

    def cleanUpAfterTrnsysExport(self):
        self.exportConnsString = ""
        self.exportInputName = "0"
        # self.exportInitialInput = -1
        self.exportEquations = []
        self._portItemsWithParent = []

    def assignIDsToUninitializedValuesAfterJsonFormatMigration(self, generator: _id.IdGenerator) -> None:
        pass


class DeleteConnectionCommand(QUndoCommand):
    def __init__(self, conn, descr):
        super().__init__(descr)
        self.conn = conn
        self.connFromPort = self.conn.fromPort
        self.connToPort = self.conn.toPort
        self.connParent = self.conn.parent

    def redo(self):
        self.conn.deleteConn()
        self.conn = None

    def undo(self):
        self.conn = Connection(self.connFromPort, self.connToPort, self.connParent)



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QUndoCommand


class HorizSegmentMoveCommand(QUndoCommand):
    def __init__(self, segment, oldX, descr):
        super(HorizSegmentMoveCommand, self).__init__(descr)
        self.oldX = oldX
        self.newX = segment.startNode.parent.scenePos().x()
        self.seg = segment

    def redo(self):
        try:
            self.seg.startNode.parent.scenePos().y()
            self.seg.endNode.parent.scenePos().y()
        except AttributeError:
            pass
        else:
            self.seg.startNode.parent.setPos(self.newX, self.seg.startNode.parent.scenePos().y())
            self.seg.endNode.parent.setPos(self.newX, self.seg.endNode.parent.scenePos().y())

    def undo(self):
        try:
            self.seg.startNode.parent.scenePos().y()
            self.seg.endNode.parent.scenePos().y()
        except AttributeError:
            pass
        else:
            self.seg.startNode.parent.setPos(self.oldX, self.seg.startNode.parent.scenePos().y())
            self.seg.endNode.parent.setPos(self.oldX, self.seg.endNode.parent.scenePos().y())



# pylint: skip-file
# type: ignore

# -*- coding: utf-8 -*-
"""
Created on Fri Sep 30 10:06:47 2016

@author: dcarbone, mbattagl
"""

import pytrnsys.rsim.runParallelTrnsys as runTrnsys

import os

if __name__ == "__main__":

    pathConfig = os.getcwd()
    configFile = "run.config"
    runTool = runTrnsys.RunParallelTrnsys(pathConfig, configFile=configFile)
